from .base import SocketClient, RpcSocketClient
from .multiplex_rpc import MultiplexRpcSocketClient
from .multiplex_streams_rpc import MultiplexStreamsRPCSocketClient
from .serial_rpc import SerialRpcSocketClient
from .streams_rpc import StreamsRPCSocketClient

__all__ = [
  'SocketClient',
  'RpcSocketClient',
  'MultiplexRpcSocketClient',
  'MultiplexStreamsRPCSocketClient',
  'SerialRpcSocketClient',
  'StreamsRPCSocketClient',
]


from dataclasses import dataclass
from datetime import datetime
from typing_extensions import TypedDict
from ..core import IndexerMixin, response_parser

class TimeResponse(TypedDict):
  iso: datetime
  epoc: float

parse_response = response_parser(TimeResponse)

@dataclass
class GetTime(IndexerMixin):
  async def get_time(
    self,
    validate: bool | None = None
  ) -> TimeResponse:
    """
    Get time

    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/utility#get-time)
    """
    r = await self.request('GET', '/v4/time')
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import TypeVar, Generic, Any, is_typeddict, TypedDict as _TypedDict
from pydantic import TypeAdapter, ConfigDict, with_config
from dataclasses import is_dataclass
from .exc import ValidationError

T = TypeVar('T')

@with_config(ConfigDict(extra='allow'))
class TypedDict(_TypedDict):
  ...

class validator(Generic[T]):

  def __init__(self, Type: type[T]):
    is_record = is_dataclass(Type) or is_typeddict(Type)
    if is_record and not hasattr(Type, '__pydantic_config__'):
      setattr(Type, '__pydantic_config__', ConfigDict(extra='allow'))
    self.adapter = TypeAdapter(Type)
    
  def json(self, data: str | bytes | bytearray) -> T:
    from pydantic import ValidationError as PydanticValidationError
    try:
      return self.adapter.validate_json(data)
    except PydanticValidationError as e:
      raise ValidationError(*e.args) from e

  def python(self, data: Any) -> T:
    from pydantic import ValidationError as PydanticValidationError
    try:
      return self.adapter.validate_python(data)
    except PydanticValidationError as e:
      raise ValidationError(*e.args) from e
    
  def __call__(self, data) -> T:
    if isinstance(data, str | bytes | bytearray):
      return self.json(data)
    else:
      return self.python(data)



from dataclasses import dataclass

from dydx_v4_client import OrderFlags
from dydx_v4_client.node.builder import TxOptions
from v4_proto.dydxprotocol.clob.tx_pb2 import MsgCancelOrder
from v4_proto.dydxprotocol.clob.order_pb2 import OrderId
from v4_proto.cosmos.tx.v1beta1.service_pb2 import BroadcastTxResponse, BroadcastMode

from dydx.core import ApiError, SHORT_BLOCK_WINDOW, STATEFUL_ORDER_TIME_WINDOW
from dydx.node.core import PrivateNodeMixin

@dataclass
class CancelOrder(PrivateNodeMixin):
  async def cancel_order(
    self, order_id: OrderId, *,
    good_til_block: int | None = None,
    good_til_block_time: int | None = None,
    mode: BroadcastMode = BroadcastMode.BROADCAST_MODE_SYNC,
    tx_options: TxOptions | None = None,
  ) -> BroadcastTxResponse:
    """Cancel an order.
    
    - `order_id`: order to cancel
    - `good_til_block`: block number to cancel the order at (defaults to the current one)
    - `good_til_block_time`: timestamp to cancel the order at (defaults to now)

    > [dYdX API docs](https://docs.dydx.xyz/node-client/private#cancel-order)
    """
    # SHORT-TERM orders required good_til_block
    if good_til_block is None and order_id.order_flags == OrderFlags.SHORT_TERM:
      latest_block = await self.node_client.latest_block()
      good_til_block = latest_block.block.header.height + SHORT_BLOCK_WINDOW
    # LONG-TERM orders required good_til_block_time
    elif good_til_block_time is None and order_id.order_flags == OrderFlags.LONG_TERM:
      latest_block = await self.node_client.latest_block()
      good_til_block_time = latest_block.block.header.time.seconds + STATEFUL_ORDER_TIME_WINDOW
    
    tx = MsgCancelOrder(order_id=order_id, good_til_block=good_til_block, good_til_block_time=good_til_block_time)
    r: BroadcastTxResponse = await self.node_client.broadcast_message(
      self.wallet, tx, mode=mode, tx_options=tx_options # type: ignore
    )
    if r.tx_response.code != 0:
      raise ApiError(r.tx_response.code, r.tx_response)
    return r


from typing_extensions import AsyncIterable, Literal
from dataclasses import dataclass
from datetime import datetime

from .api.get_fills import GetFills, Fill

MarketType = Literal['PERPETUAL', 'SPOT']


@dataclass
class GetFillsPaged(GetFills):
  async def get_fills_paged(
    self,
    *,
    address: str,
    subaccount_number: int,
    market: str | None = None,
    market_type: MarketType | None = None,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    limit: int | None = None,
    validate: bool | None = None,
  ) -> AsyncIterable[list[Fill]]:
    """Retrieves fill records for a specific subaccount on the exchange. A fill represents a trade that has been executed.

    - `address`: The wallet address that owns the account.
    - `subaccount_number`: The identifier for the specific subaccount within the wallet address.
    - `market`: The market name (e.g. `'BTC-USD'`).
    - `market_type`: The market type (`'PERPETUAL'` or `'SPOT'`). Must be provided if `market` is provided.
    - `created_before_or_at_height`: If given, fetches fills up to and including the given block height.
    - `created_before_or_at`: If given, fetches fills up to and including the given timestamp.
    - `limit`: The max. number of fills to retrieve (default: 1000, max: 1000).
    - `validate`: Whether to validate the response against the expected schema.

    > [dYdX API docs](https://docs.dydx.xyz/indexer-client/http#get-fills)
    """
    page = 1
    while True:
      response = await self.get_fills(
        address=address,
        subaccount_number=subaccount_number,
        market=market,
        market_type=market_type,
        created_before_or_at_height=created_before_or_at_height,
        created_before_or_at=created_before_or_at,
        limit=limit,
        page=page,
        validate=validate,
      )
      fills = response['fills']
      if not fills:
        break
      yield fills
      page += 1



from dataclasses import dataclass as _dataclass, field as _field
from .data import IndexerData, INDEXER_HTTP_URL
from .streams import IndexerStreams, INDEXER_WS_URL

@_dataclass
class Indexer:
  data: IndexerData = _field(default_factory=IndexerData)
  streams: IndexerStreams = _field(default_factory=IndexerStreams)

  @classmethod
  def new(cls, *, http_url: str = INDEXER_HTTP_URL, ws_url: str = INDEXER_WS_URL, validate: bool = True):
    return cls(
      data=IndexerData(url=http_url, default_validate=validate),
      streams=IndexerStreams.new(url=ws_url, validate=validate),
    )

  async def __aenter__(self):
    await self.data.__aenter__()
    await self.streams.__aenter__()
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.data.__aexit__(exc_type, exc_value, traceback)
    await self.streams.__aexit__(exc_type, exc_value, traceback)


from typing_extensions import Literal
from decimal import Decimal
from dydx.core import TypedDict

MarketType = Literal['CROSS', 'ISOLATED']
MarketStatus = Literal['ACTIVE', 'PAUSED', 'CANCEL_ONLY', 'POST_ONLY', 'INITIALIZING', 'FINAL_SETTLEMENT']

class PerpetualMarket(TypedDict):
  """Perpetual Market Object

  > [dYdX API docs](https://docs.dydx.xyz/types/perpetual_market)
  """
  clobPairId: str
  ticker: str
  status: MarketStatus
  oraclePrice: Decimal
  priceChange24H: Decimal
  volume24H: Decimal
  trades24H: int
  nextFundingRate: Decimal
  initialMarginFraction: Decimal
  maintenanceMarginFraction: Decimal
  openInterest: Decimal
  atomicResolution: int
  quantumConversionExponent: int
  tickSize: Decimal
  stepSize: Decimal
  stepBaseQuantums: int
  subticksPerTick: int
  marketType: MarketType
  openInterestLowerCap: Decimal
  openInterestUpperCap: Decimal
  baseOpenInterest: Decimal
  defaultFundingRate1H: Decimal


from typing_extensions import Sequence
from dataclasses import dataclass
from collections import defaultdict

from dydx_v4_client import OrderFlags
from v4_proto.dydxprotocol.clob.tx_pb2 import OrderBatch
from v4_proto.dydxprotocol.clob.order_pb2 import OrderId
from v4_proto.cosmos.tx.v1beta1.service_pb2 import BroadcastTxResponse

from dydx.core import ApiError, UserError, SHORT_BLOCK_WINDOW
from dydx.node.core import PrivateNodeMixin

@dataclass
class BatchCancelOrders(PrivateNodeMixin):
  async def batch_cancel_orders(
    self, order_ids: Sequence[OrderId], *,
    good_til_block: int | None = None,
  ) -> BroadcastTxResponse:
    """Cancel an order.
    
    - `order_ids`: orders to cancel
    - `good_til_block`: block number to cancel the orders at (defaults to the current one)

    > [dYdX API docs](https://docs.dydx.xyz/node-client/private#cancel-order)
    """
    subaccount_id = order_ids[0].subaccount_id
    orders_by_pair = defaultdict[int, list[int]](list)
    for order_id in order_ids:
      if order_id.order_flags != OrderFlags.SHORT_TERM:
        raise UserError('Only short-term orders can be cancelled in a batch')
      orders_by_pair[order_id.clob_pair_id].append(order_id.client_id)

    batches = [
      OrderBatch(clob_pair_id=pair, client_ids=orders)
      for pair, orders in orders_by_pair.items()
    ]

    if good_til_block is None:
      latest_block = await self.node_client.latest_block()
      good_til_block = latest_block.block.header.height + SHORT_BLOCK_WINDOW

    r: BroadcastTxResponse = await self.node_client.batch_cancel_orders(
      wallet=self.wallet, subaccount_id=subaccount_id,
      short_term_cancels=batches,
      good_til_block=good_til_block,
    )
    if r.tx_response.code != 0:
      raise ApiError(r.tx_response.code, r.tx_response)
    return r


from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class Order(TypedDict):
  id: str
  subaccountId: str
  clientId: str
  clobPairId: str
  side: Literal['BUY', 'SELL']
  size: Decimal
  totalFilled: Decimal
  price: Decimal
  type: Literal['LIMIT', 'MARKET', 'STOP_LIMIT', 'STOP_MARKET', 'TRAILING_STOP', 'TAKE_PROFIT', 'TAKE_PROFIT_MARKET', 'HARD_TRADE', 'FAILED_HARD_TRADE', 'TRANSFER_PLACEHOLDER']
  status: Literal['OPEN', 'FILLED', 'CANCELED', 'BEST_EFFORT_CANCELED', 'UNTRIGGERED', 'BEST_EFFORT_OPENED', 'PENDING']
  timeInForce: Literal['GTT', 'IOC', 'FOK']
  reduceOnly: NotRequired[bool | None]
  orderFlags: str
  goodTilBlock: NotRequired[str | None]
  goodTilBlockTime: NotRequired[datetime | None]
  createdAtHeight: NotRequired[str | None]
  clientMetadata: NotRequired[str | None]
  triggerPrice: NotRequired[Decimal | None]
  postOnly: NotRequired[bool | None]
  ticker: str
  updatedAt: NotRequired[datetime | None]
  updatedAtHeight: NotRequired[str | None]
  subaccountNumber: int
  orderRouterAddress: NotRequired[str|None]
  builderFee: NotRequired[Decimal | None]
  feePpm: NotRequired[Decimal | None]

parse_response = response_parser(Order)

@dataclass
class GetOrder(IndexerMixin):
  async def get_order(
    self,
    order_id: str,
    validate: bool | None = None
  ) -> Order:
    """
    Retrieve a single order by order id.

    - `order_id`: Order id.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-order)
    """
    r = await self.request('GET', f'/v4/orders/{order_id}')
    return parse_response(r, validate=self.validate(validate))




# Generated endpoint mixins

from .get_asset_positions import GetAssetPositions
from .get_compliance_screen import GetComplianceScreen
from .get_funding_payments_for_parent_subaccount import GetFundingPaymentsForParentSubaccount
from .get_height import GetHeight
from .get_historical_pnl import GetHistoricalPnl
from .get_megavault_historical_pnl import GetMegavaultHistoricalPnl
from .get_megavault_positions import GetMegavaultPositions
from .get_order import GetOrder
from .get_order_book import GetOrderBook
from .get_parent_asset_positions import GetParentAssetPositions
from .get_parent_fills import GetParentFills
from .get_parent_historical_pnl import GetParentHistoricalPnl
from .get_parent_subaccount import GetParentSubaccount
from .get_parent_subaccount_number_fills import GetParentSubaccountNumberFills
from .get_parent_subaccount_number_orders import GetParentSubaccountNumberOrders
from .get_parent_subaccount_number_transfers import GetParentSubaccountNumberTransfers
from .get_parent_transfers import GetParentTransfers
from .get_rewards import GetRewards
from .get_rewards_aggregated import GetRewardsAggregated
from .get_screen import GetScreen
from .get_sparklines import GetSparklines
from .get_subaccount import GetSubaccount
from .get_subaccounts import GetSubaccounts
from .get_time import GetTime
from .get_trades import GetTrades
from .get_transfers_between import GetTransfersBetween
from .get_vaults_historical_pnl import GetVaultsHistoricalPnl
from .list_orders import ListOrders
from .list_parent_orders import ListParentOrders
from .list_parent_positions import ListParentPositions



from dataclasses import dataclass
from decimal import Decimal
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class PerpetualMarket(TypedDict):
  clobPairId: str
  ticker: str
  status: Literal['ACTIVE', 'PAUSED', 'CANCEL_ONLY', 'POST_ONLY', 'INITIALIZING', 'FINAL_SETTLEMENT']
  oraclePrice: Decimal
  priceChange24H: Decimal
  volume24H: Decimal
  trades24H: int
  nextFundingRate: Decimal
  initialMarginFraction: Decimal
  maintenanceMarginFraction: Decimal
  openInterest: Decimal
  atomicResolution: int
  quantumConversionExponent: int
  tickSize: Decimal
  stepSize: Decimal
  stepBaseQuantums: int
  subticksPerTick: int
  marketType: Literal['CROSS', 'ISOLATED']
  openInterestLowerCap: NotRequired[Decimal | None]
  openInterestUpperCap: NotRequired[Decimal | None]
  baseOpenInterest: Decimal
  defaultFundingRate1H: NotRequired[Decimal | None]

class GetMarketsResponse(TypedDict):
  markets: dict[str, PerpetualMarket]

parse_response = response_parser(GetMarketsResponse)

@dataclass
class GetMarkets(IndexerMixin):
  async def get_markets(
    self,
    *,
    market: str | None = None,
    limit: int | None = None,
    validate: bool | None = None
  ) -> GetMarketsResponse:
    """
    Retrieve perpetual market metadata from the indexer. When `market` is provided, the request filters to a single market.

    - `market`: Specific market ticker to retrieve, for example `BTC-USD`.
    - `limit`: Maximum number of market entries to return.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-perpetual-markets)
    """
    params: dict[str, object] = {}
    if market is not None:
      params['market'] = market
    if limit is not None:
      params['limit'] = limit
    r = await self.request('GET', '/v4/perpetualMarkets', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import TypeVar, Any, Mapping
from dataclasses import dataclass, field
import pydantic
import httpx

from typed_core import HttpClient, ApiError
from typed_core.util import path_join

T = TypeVar('T')

def response_parser(type: type[T]):
  adapter = pydantic.TypeAdapter(type)
  def parse_response(r: httpx.Response, *, validate: bool = True) -> T:
    if r.status_code == 200:
      return adapter.validate_json(r.text) if validate else r.json()
    else:
      raise ApiError(r.status_code, r.json())
  return parse_response

INDEXER_HTTP_URL = 'https://indexer.dydx.trade/'

@dataclass(kw_only=True)
class IndexerMixin:
  url: str = INDEXER_HTTP_URL
  client: HttpClient = field(default_factory=HttpClient)
  default_validate: bool = True
  
  def validate(self, validate: bool | None = None) -> bool:
    return self.default_validate if validate is None else validate

  async def __aenter__(self):
    await self.client.__aenter__()
    return self
  
  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.client.__aexit__(exc_type, exc_value, traceback)

  async def request(
    self, method: str, path: str,
    *,
    content: httpx._types.RequestContent | None = None,
    data: httpx._types.RequestData | None = None,
    files: httpx._types.RequestFiles | None = None,
    json: Any | None = None,
    params: Mapping[str, Any] | None = None,
    headers: Mapping[str, str] | None = None,
    cookies: httpx._types.CookieTypes | None = None,
    auth: httpx._types.AuthTypes | httpx._client.UseClientDefault | None = httpx.USE_CLIENT_DEFAULT,
    follow_redirects: bool | httpx._client.UseClientDefault = httpx.USE_CLIENT_DEFAULT,
    timeout: httpx._types.TimeoutTypes | httpx._client.UseClientDefault = httpx.USE_CLIENT_DEFAULT,
    extensions: httpx._types.RequestExtensions | None = None,
  ):
    url = path_join(self.url, path)
    return await self.client.request(
      method, url, params=params, cookies=cookies, json=json,
      content=content, data=data, files=files, auth=auth, follow_redirects=follow_redirects,
      timeout=timeout, extensions=extensions,
      headers=headers,
    )


from typing_extensions import Any

class Error(Exception):
  def __str__(self):
    args = self.args[0] if len(self.args) == 1 else ', '.join(self.args)
    return f'{self.__class__.__name__}({args})'

class NetworkError(Error):
  def __str__(self):
    return super().__str__()

class ValidationError(Error):
  def __str__(self):
    return super().__str__()

class UserError(Error):
  def __str__(self):
    return super().__str__()

class AuthError(Error):
  def __str__(self):
    return super().__str__()

class ApiError(Error):
  def __init__(self, status: Any | None, result, *args):
    super().__init__(result, *args)
    self.status = status
    self.result = result

  def __str__(self):
    return f'{self.__class__.__name__}(status: {self.status}, detail: {self.result})'


from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class Candle(TypedDict):
  startedAt: datetime
  ticker: str
  resolution: Literal['1MIN', '5MINS', '15MINS', '30MINS', '1HOUR', '4HOURS', '1DAY']
  low: Decimal
  high: Decimal
  open: Decimal
  close: Decimal
  baseTokenVolume: Decimal
  usdVolume: Decimal
  trades: int
  startingOpenInterest: Decimal
  orderbookMidPriceOpen: NotRequired[Decimal | None]
  orderbookMidPriceClose: NotRequired[Decimal | None]

class CandlesResponse(TypedDict):
  candles: list[Candle]

parse_response = response_parser(CandlesResponse)

@dataclass
class GetCandles(IndexerMixin):
  async def get_candles(
    self,
    market: str,
    *,
    resolution: Literal[
      '1MIN', '5MINS', '15MINS', '30MINS', '1HOUR', '4HOURS', '1DAY'
    ],
    from_iso: datetime | None = None,
    to_iso: datetime | None = None,
    limit: int | None = None,
    validate: bool | None = None
  ) -> CandlesResponse:
    """
    Retrieve candle data for a perpetual market.

    - `market`: Perpetual market ticker.
    - `resolution`: Candle resolution.
    - `from_iso`: Start timestamp in ISO 8601 format.
    - `to_iso`: End timestamp in ISO 8601 format.
    - `limit`: Maximum number of candles to return.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-candles)
    """
    params: dict[str, object] = {
      'resolution': resolution,
    }
    if from_iso is not None:
      params['fromISO'] = ts.dump(from_iso)
    if to_iso is not None:
      params['toISO'] = ts.dump(to_iso)
    if limit is not None:
      params['limit'] = limit
    r = await self.request('GET', f'/v4/candles/perpetualMarkets/{market}', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass

from grpc._channel import _InactiveRpcError

from v4_proto.dydxprotocol.clob.clob_pair_pb2 import ClobPair
from dydx.core import ApiError
from dydx.node.core import PublicNodeMixin

@dataclass
class GetClobPair(PublicNodeMixin):
  async def get_clob_pair(self, id: int) -> ClobPair:
    """
    Fetches the order book pair identified by a given ID, allowing users to retrieve detailed information about a specific trading pair within a Central Limit Order Book (CLOB) system.

    - `id`: The id of the CLOB pair.

    > [dYdX API docs](https://docs.dydx.xyz/node-client/public#get-clob-pair)
    """
    try:
      return await self.node_client.get_clob_pair(id)
    except _InactiveRpcError as e:
      raise ApiError(e._state.code, e._state.details)


from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing_extensions import Literal, TypedDict
from ..core import IndexerMixin, response_parser

class PnlTick(TypedDict):
  blockHeight: str
  blockTime: datetime
  createdAt: datetime
  equity: Decimal
  totalPnl: Decimal
  netTransfer: Decimal

class VaultHistoricalPnl(TypedDict):
  ticker: str
  historicalPnl: PnlTick

parse_response = response_parser(list[VaultHistoricalPnl])

@dataclass
class GetVaultsHistoricalPnl(IndexerMixin):
  async def get_vaults_historical_pnl(
    self,
    *,
    resolution: Literal['hour', 'day'],
    validate: bool | None = None
  ) -> list[VaultHistoricalPnl]:
    """
    Get vaults historical pnl

    - `resolution`: PnL tick resolution.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/vaults#get-vaults-historical-pnl)
    """
    params: dict[str, object] = {
      'resolution': resolution,
    }
    r = await self.request('GET', '/v4/vault/v1/vaults/historicalPnl', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import Literal, Any, NotRequired, TypedDict
from dataclasses import dataclass, field
import asyncio
import logging
import orjson
import pydantic

from typed_core import LogicError, BadRequest
from typed_core.ws.serial_streams import SerialStreams, Subscription

logger = logging.getLogger('dydx.indexer.streams')

INDEXER_WS_URL = 'wss://indexer.dydx.trade/v4/ws'

class BaseMessage(TypedDict):
  connection_id: str
  message_id: int

class Connected(BaseMessage):
  type: Literal['connected']

class Channel(TypedDict):
  channel: str
  id: NotRequired[str]

class Subscribed(BaseMessage, Channel):
  type: Literal['subscribed']
  contents: Any

class Unsubscribed(BaseMessage, Channel):
  type: Literal['unsubscribed']

class Error(BaseMessage):
  type: Literal['error']

class Notification(BaseMessage, Channel):
  type: Literal['channel_data', 'channel_batch_data']
  version: str
  contents: Any

Msg = Connected | Subscribed | Unsubscribed | Error | Notification
MsgT: type[Msg] = Msg # type: ignore

msg_adapter = pydantic.TypeAdapter(MsgT)

class Params(TypedDict, total=False):
  batched: bool

def parse_channel_id(channel_id: str) -> tuple[str, str|None]:
  if ':' in channel_id:
    channel, id = channel_id.split(':')
    return channel, id
  else:
    return channel_id, None

def channel_id(msg: Channel) -> str:
  out = msg['channel']
  if (id := msg.get('id')) is not None and id != msg['channel']: # yes that happens, it's a dydx bug
    out += f':{id}'
  return out

@dataclass
class StreamsClient(SerialStreams[Notification, Params, Subscribed, Unsubscribed]):
  url: str = INDEXER_WS_URL
  replies: asyncio.Queue[Error | Subscribed | Unsubscribed] = field(default_factory=asyncio.Queue)

  def parse_msg(self, msg: str | bytes) -> Subscription | None:
    obj = msg_adapter.validate_json(msg)
    match obj['type']:
      case 'subscribed' | 'unsubscribed' | 'error':
        self.replies.put_nowait(obj)
      case 'channel_data' | 'channel_batch_data':
        channel = channel_id(obj)
        return {'channel': channel, 'notification': obj}

  async def send(self, msg):
    ws = await self.ws
    await ws.send(orjson.dumps(msg), text=True)

  async def request_subscription(self, channel: str, params: Params | None = None) -> Subscribed:
    channel, id = parse_channel_id(channel)
    msg: dict = {
      'type': 'subscribe',
      'channel': channel,
    }
    if (params or {}).get('batched', False):
      msg['batched'] = True
    if id is not None:
      msg['id'] = id
    await self.send(msg)
    reply = await self.replies.get()
    if reply['type'] == 'error':
      raise BadRequest(reply)
    elif reply['type'] != 'subscribed':
      raise LogicError(f'Unexpected response type: {reply}')
    return reply

  async def request_unsubscription(self, channel: str, params: Params | None = None) -> Unsubscribed:
    channel, id = parse_channel_id(channel)
    msg = {
      'type': 'unsubscribe',
      'channel': channel,
    }
    if id is not None:
      msg['id'] = id
    await self.send(msg)
    reply = await self.replies.get()
    if reply['type'] == 'error':
      raise BadRequest(reply)
    if reply['type'] != 'unsubscribed':
      raise LogicError(f'Unexpected response type: {reply}')
    return reply

@dataclass(kw_only=True)
class StreamsMixin:
  client: StreamsClient = field(default_factory=StreamsClient)
  default_validate: bool = True

  def validate(self, validate: bool | None = None) -> bool:
    return self.default_validate if validate is None else validate

  async def __aenter__(self):
    await self.client.__aenter__()
    return self
  
  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.client.__aexit__(exc_type, exc_value, traceback)

  @classmethod
  def new(cls, url: str = INDEXER_WS_URL, *, validate: bool = True):
    return cls(client=StreamsClient(url=url), default_validate=validate)


from dydx.core import TypedDict
from .positions import PerpetualPosition, AssetPosition

class Account(TypedDict):
  """Account Object

  > [dYdX API docs](https://docs.dydx.xyz/types/account)
  """
  address: str
  subaccountNumber: int


class Subaccount(TypedDict):
  address: str
  subaccountNumber: int
  equity: str
  freeCollateral: str
  openPerpetualPositions: dict[str, PerpetualPosition]
  assetPositions: dict[str, AssetPosition]
  marginEnabled: bool
  updatedAtHeight: str
  latestProcessedBlockHeight: str


from dataclasses import dataclass
from decimal import Decimal
from typing_extensions import Literal, TypedDict
from ..core import IndexerMixin, response_parser

class AssetPosition(TypedDict):
  size: Decimal
  symbol: str
  side: Literal['LONG', 'SHORT']
  assetId: str
  subaccountNumber: int

parse_response = response_parser(list[AssetPosition])

@dataclass
class GetParentAssetPositions(IndexerMixin):
  async def get_parent_asset_positions(
    self,
    *,
    address: str,
    parent_subaccount_number: int,
    validate: bool | None = None
  ) -> list[AssetPosition]:
    """
    Get parent asset positions

    - `address`: Wallet address that owns the account.
    - `parent_subaccount_number`: Parent subaccount number.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-parent-asset-positions)
    """
    params: dict[str, object] = {
      'address': address,
      'parentSubaccountNumber': parent_subaccount_number,
    }
    r = await self.request('GET', '/v4/assetPositions/parentSubaccountNumber', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, TypedDict
from ..core import IndexerMixin, response_parser

class HistoricalTradingRewardAggregation(TypedDict):
  tradingReward: Decimal
  startedAtHeight: str
  startedAt: datetime
  endedAtHeight: str
  endedAt: datetime

parse_response = response_parser(list[HistoricalTradingRewardAggregation])

@dataclass
class GetRewardsAggregated(IndexerMixin):
  async def get_rewards_aggregated(
    self,
    address: str,
    *,
    period: Literal['DAILY', 'WEEKLY', 'MONTHLY'],
    limit: int | None = None,
    starting_before_or_at: datetime | None = None,
    starting_before_or_at_height: int | None = None,
    validate: bool | None = None
  ) -> list[HistoricalTradingRewardAggregation]:
    """
    Get aggregated trading rewards

    - `address`: Wallet address that owns the account.
    - `period`: Aggregation period.
    - `limit`: Maximum number of results to return.
    - `starting_before_or_at`: Only include aggregations starting at or before this timestamp.
    - `starting_before_or_at_height`: Only include aggregations starting at or before this block height.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-rewards-aggregated)
    """
    params: dict[str, object] = {
      'period': period,
    }
    if limit is not None:
      params['limit'] = limit
    if starting_before_or_at is not None:
      params['startingBeforeOrAt'] = ts.dump(starting_before_or_at)
    if starting_before_or_at_height is not None:
      params['startingBeforeOrAtHeight'] = starting_before_or_at_height
    r = await self.request('GET', f'/v4/historicalTradingRewardAggregations/{address}', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import AsyncIterable, Literal, TypedDict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
import pydantic

from typed_core.ws import Stream
from ..core import StreamsMixin, Unsubscribed

class Trade(TypedDict):
  id: str
  side: Literal['BUY', 'SELL']
  size: Decimal
  price: Decimal
  type: Literal['LIMIT', 'LIQUIDATED', 'DELEVERAGED']
  createdAt: datetime
  createdAtHeight: str

class TradeUpdate(TypedDict):
  id: str
  createdAt: datetime
  side: Literal['BUY', 'SELL']
  price: Decimal
  size: Decimal
  type: Literal['LIMIT', 'LIQUIDATED', 'DELEVERAGED']

class Reply(TypedDict):
  trades: list[Trade]

class Notification(TypedDict):
  trades: list[TradeUpdate]

reply_adapter = pydantic.TypeAdapter(Reply)
notification_adapter = pydantic.TypeAdapter(Notification)

@dataclass
class Trades(StreamsMixin):
  async def trades(
    self, *, id: str, batched: bool = True, validate: bool | None = None,
  ) -> Stream[Notification, Reply, Unsubscribed]:
    """
    Subscribe to the indexer trades feed for a market.

    - `id`: Market ticker.
    - `batched`: Reduce incoming messages by batching contents.
    - `validate`: Whether to validate reply and update payloads against the generated schemas.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/websockets#trades)
    """
    stream = await self.client.subscribe(f'v4_trades:{id}', {'batched': batched})

    async def parsed_stream() -> AsyncIterable[Notification]:
      async for msg in stream:
        data = msg['contents']
        msgs: list[Notification] = data if batched else [data]
        for d in msgs:
          yield notification_adapter.validate_python(d) if self.validate(validate) else d

    c = stream.reply['contents']
    reply: Reply = reply_adapter.validate_python(c) if self.validate(validate) else c
    return Stream(reply, parsed_stream(), stream.unsubscribe)




from .validation import validator, TypedDict
from .util import timestamp, round2tick, trunc2tick, filter_kwargs, path_join, getenv
from .exc import Error, NetworkError, UserError, ValidationError, AuthError, ApiError
from .http import HttpClient
from .constants import SHORT_BLOCK_WINDOW, STATEFUL_ORDER_TIME_WINDOW

__all__ = [
  'validator', 'TypedDict',
  'timestamp', 'round2tick', 'trunc2tick', 'filter_kwargs', 'path_join', 'getenv',
  'Error', 'NetworkError', 'UserError', 'ValidationError', 'AuthError', 'ApiError',
  'HttpClient',
  'SHORT_BLOCK_WINDOW', 'STATEFUL_ORDER_TIME_WINDOW',
]


from typing_extensions import NotRequired

from dydx.core import TypedDict
from .enums import OrderSide, OrderStatus, OrderType, TimeInForce

class OrderState(TypedDict):
  """Order Response Object

  > [dYdX API docs](https://docs.dydx.xyz/types/order_response_object)
  """
  id: str
  subaccountId: str
  clientId: str
  clobPairId: str
  side: OrderSide
  size: str
  totalFilled: str
  price: str
  type: OrderType
  status: OrderStatus
  timeInForce: TimeInForce
  reduceOnly: NotRequired[bool|None]
  orderFlags: str
  goodTilBlock: NotRequired[str|None]
  createdAtHeight: NotRequired[str|None]
  clientMetadata: NotRequired[str|None]
  updatedAt: NotRequired[str|None]
  updatedAtHeight: NotRequired[str|None]
  orderRouterAddress: NotRequired[str|None]
  postOnly: NotRequired[bool|None]
  ticker: str
  subaccountNumber: int



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class AssetPosition(TypedDict):
  size: Decimal
  symbol: str
  side: Literal['LONG', 'SHORT']
  assetId: str
  subaccountNumber: int

class PerpetualPosition(TypedDict):
  market: str
  status: Literal['OPEN', 'CLOSED', 'LIQUIDATED']
  side: Literal['LONG', 'SHORT']
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal | None]
  realizedPnl: NotRequired[Decimal | None]
  unrealizedPnl: NotRequired[Decimal | None]
  createdAt: datetime
  createdAtHeight: str
  closedAt: NotRequired[datetime | None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

AssetPositionsMap = dict[str, AssetPosition]

PerpetualPositionsMap = dict[str, PerpetualPosition]

class Subaccount(TypedDict):
  address: str
  subaccountNumber: int
  equity: Decimal
  freeCollateral: Decimal
  openPerpetualPositions: PerpetualPositionsMap
  assetPositions: AssetPositionsMap
  marginEnabled: bool
  updatedAtHeight: str
  latestProcessedBlockHeight: str

class GetSubaccountResponse(TypedDict):
  subaccount: Subaccount

parse_response = response_parser(GetSubaccountResponse)

@dataclass
class GetSubaccount(IndexerMixin):
  async def get_subaccount(
    self,
    address: str,
    subaccount_number: int,
    validate: bool | None = None
  ) -> GetSubaccountResponse:
    """
    Retrieve a single subaccount.

    - `address`: Account address.
    - `subaccount_number`: Subaccount number.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-subaccount)
    """
    r = await self.request('GET', f'/v4/addresses/{address}/subaccountNumber/{subaccount_number}')
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import Any, TypeVar, Generic
from abc import abstractmethod
from dataclasses import dataclass, field
import asyncio

from .base import RpcSocketClient

M = TypeVar('M', default=Any)

@dataclass
class SerialRpcSocketClient(RpcSocketClient[M, str|bytes], Generic[M]):
  """Serial request/response socket client. It uses a lock to serialize requests and responses."""
  lock: asyncio.Lock = field(default_factory=asyncio.Lock)
  replies: asyncio.Queue[str|bytes] = field(default_factory=asyncio.Queue)

  def on_msg(self, msg: str | bytes):
    self.replies.put_nowait(msg)

  @abstractmethod
  async def send(self, msg: M):
    ...

  async def request(self, msg: M) -> str | bytes:
    async with self.lock:
      await self.send(msg)
      res = await self.replies.get()
      return res



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, TypedDict
from ..core import IndexerMixin, response_parser

class FundingPayment(TypedDict):
  createdAt: datetime
  createdAtHeight: str
  perpetualId: str
  ticker: str
  oraclePrice: Decimal
  size: Decimal
  side: Literal['LONG', 'SHORT']
  rate: Decimal
  payment: Decimal
  subaccountNumber: str
  fundingIndex: Decimal

class FundingPaymentsResponse(TypedDict):
  fundingPayments: list[FundingPayment]

parse_response = response_parser(FundingPaymentsResponse)

@dataclass
class GetFundingPaymentsForParentSubaccount(IndexerMixin):
  async def get_funding_payments_for_parent_subaccount(
    self,
    *,
    address: str,
    parent_subaccount_number: int,
    limit: int | None = None,
    after_or_at: datetime | None = None,
    page: int | None = None,
    validate: bool | None = None
  ) -> FundingPaymentsResponse:
    """
    Get funding payments for parent subaccount

    - `address`: Wallet address that owns the account.
    - `parent_subaccount_number`: Parent subaccount number.
    - `limit`: Maximum number of results to return.
    - `after_or_at`: Only include payments created at or after this timestamp.
    - `page`: Page number for paginated results.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-funding-payments-for-parent-subaccount)
    """
    params: dict[str, object] = {
      'address': address,
      'parentSubaccountNumber': parent_subaccount_number,
    }
    if limit is not None:
      params['limit'] = limit
    if after_or_at is not None:
      params['afterOrAt'] = ts.dump(after_or_at)
    if page is not None:
      params['page'] = page
    r = await self.request('GET', '/v4/fundingPayments/parentSubaccount', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, TypedDict
from ..core import IndexerMixin, response_parser

class AssetPosition(TypedDict):
  size: Decimal
  symbol: str
  side: Literal['LONG', 'SHORT']
  assetId: str
  subaccountNumber: int

parse_response = response_parser(list[AssetPosition])

@dataclass
class GetAssetPositions(IndexerMixin):
  async def get_asset_positions(
    self,
    *,
    address: str,
    subaccount_number: int,
    status: Literal['OPEN', 'CLOSED', 'LIQUIDATED'] | None = None,
    limit: int | None = None,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    validate: bool | None = None
  ) -> list[AssetPosition]:
    """
    Get asset positions

    - `address`: Wallet address that owns the account.
    - `subaccount_number`: Subaccount number.
    - `status`: Position status filter.
    - `limit`: Maximum number of results to return.
    - `created_before_or_at_height`: Restrict results to entries created at or before a specific block height.
    - `created_before_or_at`: Restrict results to entries created at or before a specific timestamp.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-asset-positions)
    """
    params: dict[str, object] = {
      'address': address,
      'subaccountNumber': subaccount_number,
    }
    if status is not None:
      params['status'] = status
    if limit is not None:
      params['limit'] = limit
    if created_before_or_at_height is not None:
      params['createdBeforeOrAtHeight'] = created_before_or_at_height
    if created_before_or_at is not None:
      params['createdBeforeOrAt'] = ts.dump(created_before_or_at)
    r = await self.request('GET', '/v4/assetPositions', params=params)
    return parse_response(r, validate=self.validate(validate))




import os
from dataclasses import dataclass, field
from dydx_v4_client.node.client import NodeClient
from dydx_v4_client.key_pair import KeyPair
from dydx_v4_client.wallet import Wallet
from dydx_v4_client.network import make_mainnet

from dydx.indexer import INDEXER_HTTP_URL, INDEXER_WS_URL

OEGS_GRPC_URL = 'oegs.dydx.trade:443'

@dataclass
class PublicNodeMixin:
  node_client: NodeClient = field(kw_only=True)
  
  @classmethod
  async def connect(
    cls, *, url: str = OEGS_GRPC_URL,
    rest_indexer: str = INDEXER_HTTP_URL,
    websocket_indexer: str = INDEXER_WS_URL,
  ):
    config = make_mainnet( 
      node_url=url,
      rest_indexer=rest_indexer, 
      websocket_indexer=websocket_indexer, 
    ).node
    node: NodeClient = await NodeClient.connect(config)
    return cls(node_client=node)

@dataclass
class PrivateNodeMixin:
  address: str = field(kw_only=True)
  node_client: NodeClient = field(kw_only=True)
  wallet: Wallet = field(kw_only=True)
  
  @classmethod
  async def connect(
    cls, mnemonic: str | None = None, *, url: str = OEGS_GRPC_URL,
    rest_indexer: str = INDEXER_HTTP_URL,
    websocket_indexer: str = INDEXER_WS_URL,
  ):
    if mnemonic is None:
      mnemonic = os.environ['DYDX_MNEMONIC']
    pair = KeyPair.from_mnemonic(mnemonic)
    address = Wallet(pair, 0, 0).address
    config = make_mainnet( 
      node_url=url,
      rest_indexer=rest_indexer, 
      websocket_indexer=websocket_indexer, 
    ).node
    node: NodeClient = await NodeClient.connect(config)
    wallet = await Wallet.from_mnemonic(node, mnemonic, address)
    return cls(node_client=node, wallet=wallet, address=address)


from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class AssetPosition(TypedDict):
  size: Decimal
  symbol: str
  side: Literal['LONG', 'SHORT']
  assetId: str
  subaccountNumber: int

class PerpetualPosition(TypedDict):
  market: str
  status: Literal['OPEN', 'CLOSED', 'LIQUIDATED']
  side: Literal['LONG', 'SHORT']
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal | None]
  realizedPnl: NotRequired[Decimal | None]
  unrealizedPnl: NotRequired[Decimal | None]
  createdAt: datetime
  createdAtHeight: str
  closedAt: NotRequired[datetime | None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

class VaultPosition(TypedDict):
  ticker: str
  assetPosition: AssetPosition
  perpetualPosition: PerpetualPosition
  equity: Decimal

parse_response = response_parser(list[VaultPosition])

@dataclass
class GetMegavaultPositions(IndexerMixin):
  async def get_megavault_positions(
    self,
    validate: bool | None = None
  ) -> list[VaultPosition]:
    """
    Get megavault positions

    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/vaults#get-megavault-positions)
    """
    r = await self.request('GET', '/v4/vault/v1/megavault/positions')
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import AsyncIterable, TypedDict
from dataclasses import dataclass
from datetime import datetime
import pydantic

from typed_core.ws import Stream
from ..core import StreamsMixin, Unsubscribed

class Reply(TypedDict):
  height: str
  time: datetime

class Notification(TypedDict):
  blockHeight: str
  time: datetime

reply_adapter = pydantic.TypeAdapter(Reply)
notification_adapter = pydantic.TypeAdapter(Notification)

@dataclass
class BlockHeight(StreamsMixin):
  async def block_height(
    self, *, batched: bool = True, validate: bool | None = None
  ) -> Stream[Notification, Reply, Unsubscribed]:
    """
    Subscribe to the indexer block height feed.

    - `batched`: Reduce incoming messages by batching contents.
    - `validate`: Whether to validate reply and update payloads against the generated schemas.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/websockets#block-height)
    """
    stream = await self.client.subscribe('v4_block_height', {'batched': batched})

    async def parsed_stream() -> AsyncIterable[Notification]:
      async for msg in stream:
        data = msg['contents']
        msgs: list[Notification] = data if batched else [data]
        for d in msgs:
          yield notification_adapter.validate_python(d) if self.validate(validate) else d

    c = stream.reply['contents']
    reply: Reply = reply_adapter.validate_python(c) if self.validate(validate) else c
    return Stream(reply, parsed_stream(), stream.unsubscribe)




from typing_extensions import Literal, TypedDict, AsyncIterable, TypeVar, Generic, Any
from abc import abstractmethod
from dataclasses import dataclass, field
from collections import defaultdict
import asyncio

from .base import RpcSocketClient

M = TypeVar('M', default=Any)
R = TypeVar('R', default=Any)
S = TypeVar('S', default=Any)

class Response(TypedDict, Generic[R]):
  kind: Literal['response']
  id: int
  response: R

class Subscription(TypedDict, Generic[S]):
  kind: Literal['subscription']
  channel: str
  data: S

Message = Response[R] | Subscription[S]

@dataclass
class MultiplexStreamsRPCSocketClient(RpcSocketClient[M, R], Generic[M, R, S]):
  """Multiplexed request/response and streams socket client. It uses IDs to identify requests and responses. It also supports subscription to multiple channels."""
  replies: dict[int, asyncio.Future[R]] = field(default_factory=dict, init=False, repr=False)
  counter: int = field(default=0, init=False, repr=False)
  subscribers: dict[str, list[asyncio.Queue[S]]] = field(default_factory=lambda: defaultdict(list), init=False, repr=False)

  async def request(self, msg: M) -> R:
    id = self.counter
    self.counter += 1
    while True:
      self.replies[id] = asyncio.Future()
      await self.send(id, msg)
      res = await self.wait_with_listener(self.replies[id])
      del self.replies[id]
      return res
    
  @abstractmethod
  async def req_subscription(self, channel: str):
    ...

  @abstractmethod
  async def req_unsubscription(self, channel: str):
    ...

  @abstractmethod
  def parse_msg(self, msg: str | bytes) -> Message[R, S]:
    ...

  def on_msg(self, msg: str | bytes):
    res = self.parse_msg(msg)
    if res['kind'] == 'response':
      self.replies[res['id']].set_result(res['response'])
    else:
      for q in self.subscribers[res['channel']]:
        q.put_nowait(res['data'])

  @abstractmethod
  async def send(self, id: int, msg: M):
    ...

  async def subscribe(self, channel: str) -> AsyncIterable[S]:
    q = asyncio.Queue()
    self.subscribers[channel].append(q)
    await self.req_subscription(channel)
    while True:
      yield await self.wait_with_listener(q.get())

  async def unsubscribe(self, channel: str):
    del self.subscribers[channel]
    await self.req_unsubscription(channel)



from dataclasses import dataclass as _dataclass

from .get_clob_pair import GetClobPair
from .get_price import GetPrice
from .get_user_fee_tier import GetUserFeeTier

@_dataclass
class PublicNode(GetClobPair, GetPrice, GetUserFeeTier):
  ...


from dataclasses import dataclass
from decimal import Decimal

from grpc._channel import _InactiveRpcError

from dydx.core import ApiError
from dydx.node.core import PublicNodeMixin

@dataclass
class GetPrice(PublicNodeMixin):
  async def get_price(self, id: int):
    """
    Retrieve the current market price for a specified market, identified by its market ID.

    - `id`: The id of the CLOB pair.

    > [dYdX API docs](https://docs.dydx.xyz/node-client/public#get-price)
    """
    try:
      r = await self.node_client.get_price(id)
      return Decimal(r.price) * Decimal(f'1e{r.exponent}')
    except _InactiveRpcError as e:
      raise ApiError(e._state.code, e._state.details)


from dataclasses import dataclass as _dataclass

from .batch_cancel_orders import BatchCancelOrders
from .cancel_order import CancelOrder
from .place_order import PlaceOrder

@_dataclass
class PrivateNode(BatchCancelOrders, CancelOrder, PlaceOrder):
  ...


from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class Order(TypedDict):
  id: str
  subaccountId: str
  clientId: str
  clobPairId: str
  side: Literal['BUY', 'SELL']
  size: Decimal
  totalFilled: Decimal
  price: Decimal
  type: Literal['LIMIT', 'MARKET', 'STOP_LIMIT', 'STOP_MARKET', 'TRAILING_STOP', 'TAKE_PROFIT', 'TAKE_PROFIT_MARKET', 'HARD_TRADE', 'FAILED_HARD_TRADE', 'TRANSFER_PLACEHOLDER']
  status: Literal['OPEN', 'FILLED', 'CANCELED', 'BEST_EFFORT_CANCELED', 'UNTRIGGERED', 'BEST_EFFORT_OPENED', 'PENDING']
  timeInForce: Literal['GTT', 'IOC', 'FOK']
  reduceOnly: NotRequired[bool | None]
  orderFlags: str
  goodTilBlock: NotRequired[str | None]
  goodTilBlockTime: NotRequired[datetime | None]
  createdAtHeight: NotRequired[str | None]
  clientMetadata: NotRequired[str | None]
  triggerPrice: NotRequired[Decimal | None]
  postOnly: NotRequired[bool | None]
  ticker: str
  updatedAt: NotRequired[datetime | None]
  updatedAtHeight: NotRequired[str | None]
  subaccountNumber: int
  orderRouterAddress: NotRequired[str|None]
  builderFee: NotRequired[Decimal | None]
  feePpm: NotRequired[Decimal | None]

parse_response = response_parser(list[Order])

@dataclass
class ListParentOrders(IndexerMixin):
  async def list_parent_orders(
    self,
    *,
    address: str,
    parent_subaccount_number: int,
    limit: int | None = None,
    ticker: str | None = None,
    side: Literal['BUY', 'SELL'] | None = None,
    status: Literal[
      'OPEN', 'FILLED', 'CANCELED', 'BEST_EFFORT_CANCELED', 'UNTRIGGERED', 'BEST_EFFORT_OPENED', 'PENDING'
    ] | None = None,
    type: Literal[
      'LIMIT', 'MARKET', 'STOP_LIMIT', 'STOP_MARKET', 'TRAILING_STOP', 'TAKE_PROFIT', 'TAKE_PROFIT_MARKET', 'HARD_TRADE', 'FAILED_HARD_TRADE', 'TRANSFER_PLACEHOLDER'
    ] | None = None,
    good_til_block_before_or_at: int | None = None,
    good_til_block_time_before_or_at: datetime | None = None,
    return_latest_orders: bool | None = None,
    validate: bool | None = None
  ) -> list[Order]:
    """
    List parent orders

    - `address`: Wallet address that owns the account.
    - `parent_subaccount_number`: Parent subaccount number.
    - `limit`: Maximum number of results to return.
    - `ticker`: Ticker filter.
    - `side`: Order side filter.
    - `status`: Order status filter.
    - `type`: Order type filter.
    - `good_til_block_before_or_at`: Latest good-til-block to include.
    - `good_til_block_time_before_or_at`: Latest good-til-block time to include.
    - `return_latest_orders`: Whether to return only the latest orders.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#list-parent-orders)
    """
    params: dict[str, object] = {
      'address': address,
      'parentSubaccountNumber': parent_subaccount_number,
    }
    if limit is not None:
      params['limit'] = limit
    if ticker is not None:
      params['ticker'] = ticker
    if side is not None:
      params['side'] = side
    if status is not None:
      params['status'] = status
    if type is not None:
      params['type'] = type
    if good_til_block_before_or_at is not None:
      params['goodTilBlockBeforeOrAt'] = good_til_block_before_or_at
    if good_til_block_time_before_or_at is not None:
      params['goodTilBlockTimeBeforeOrAt'] = ts.dump(good_til_block_time_before_or_at)
    if return_latest_orders is not None:
      params['returnLatestOrders'] = return_latest_orders
    r = await self.request('GET', '/v4/orders/parentSubaccountNumber', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import AsyncIterable, Literal, Sequence
from dataclasses import dataclass
from datetime import datetime

from dydx.core import timestamp as ts
from .api.get_candles import GetCandles, Candle

Resolution = Literal['1MIN', '5MINS', '15MINS', '30MINS', '1HOUR', '4HOURS', '1DAY']

@dataclass
class GetCandlesPaged(GetCandles):
  async def get_candles_paged(
    self,
    market: str,
    resolution: Resolution,
    *,
    start: datetime | None = None,
    end: datetime | None = None,
    limit: int | None = None,
  ) -> AsyncIterable[Sequence[Candle]]:
    """Retrieves candle data for a specific perpetual market, paging as needed.

    - `market`: The market ticker (e.g. `'BTC-USD'`).
    - `resolution`: The resolution of the candles.
    - `start`: If given, fetches candles starting from the given timestamp.
    - `end`: If given, fetches candles up to and including the given timestamp.
    - `limit`: The max. number of candles to retrieve per request (default: 1000, max: 1000).
    """
    last_time = end
    while True:
      response = await self.get_candles(
        market,
        resolution=resolution,
        from_iso=start,
        to_iso=last_time,
        limit=limit,
      )
      candles = response['candles']
      if not candles:
        break
      yield candles
      last_time = candles[-1]['startedAt']



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class PerpetualPosition(TypedDict):
  market: str
  status: Literal['OPEN', 'CLOSED', 'LIQUIDATED']
  side: Literal['LONG', 'SHORT']
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal | None]
  realizedPnl: NotRequired[Decimal | None]
  unrealizedPnl: NotRequired[Decimal | None]
  createdAt: datetime
  createdAtHeight: str
  closedAt: NotRequired[datetime | None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

parse_response = response_parser(list[PerpetualPosition])

@dataclass
class ListParentPositions(IndexerMixin):
  async def list_parent_positions(
    self,
    *,
    address: str,
    parent_subaccount_number: int,
    limit: int | None = None,
    validate: bool | None = None
  ) -> list[PerpetualPosition]:
    """
    List parent positions

    - `address`: Wallet address that owns the account.
    - `parent_subaccount_number`: Parent subaccount number.
    - `limit`: Maximum number of results to return.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#list-parent-positions)
    """
    params: dict[str, object] = {
      'address': address,
      'parentSubaccountNumber': parent_subaccount_number,
    }
    if limit is not None:
      params['limit'] = limit
    r = await self.request('GET', '/v4/perpetualPositions/parentSubaccountNumber', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import AsyncIterable, Sequence
from dataclasses import dataclass
from datetime import datetime

from .api.get_historical_funding import GetHistoricalFunding, Funding


@dataclass
class GetHistoricalFundingPaged(GetHistoricalFunding):
  async def get_historical_funding_paged(
    self,
    market: str,
    *,
    effective_before_or_at: datetime | None = None,
    effective_before_or_at_height: int | None = None,
    limit: int | None = None,
  ) -> AsyncIterable[Sequence[Funding]]:
    """Retrieves historical funding rates for a specific perpetual market, automatically paginating.

    - `market`: The market ticker (e.g. `'BTC-USD'`).
    - `effective_before_or_at`: If given, fetches funding rates up to and including the given timestamp.
    - `effective_before_or_at_height`: If given, fetches funding rates up to and including the given block height.
    - `limit`: The max. number of candles to retrieve (default: 1000, max: 1000).
    """
    last_block = effective_before_or_at_height
    while True:
      response = await self.get_historical_funding(
        market,
        effective_before_or_at=effective_before_or_at,
        effective_before_or_at_height=last_block,
        limit=limit,
      )
      historical_funding = response['historicalFunding']
      if not historical_funding:
        break
      new_last_block = int(historical_funding[-1]['effectiveAtHeight'])
      if new_last_block == last_block:
        break
      last_block = new_last_block
      yield historical_funding



from typing_extensions import AsyncIterable
from dataclasses import dataclass
from datetime import datetime

from .api.get_funding_payments import GetFundingPayments, FundingPayment

@dataclass
class GetFundingPaymentsPaged(GetFundingPayments):
  async def get_funding_payments_paged(
    self,
    *,
    address: str,
    subaccount_number: int,
    ticker: str | None = None,
    after_or_at: datetime | None = None,
    limit: int | None = None,
    validate: bool | None = None,
  ) -> AsyncIterable[list[FundingPayment]]:
    """Retrieves funding payment history for a specific subaccount. Funding payments are periodic settlements that occur between long and short positions based on the funding rate.

    - `address`: The wallet address that owns the account.
    - `subaccount_number`: The identifier for the specific subaccount within the wallet address.
    - `ticker`: The market ticker (e.g. `'BTC-USD'`).
    - `after_or_at`: If given, fetches funding payments starting from the given timestamp.
    - `limit`: The max. number of funding payments to retrieve.
    - `validate`: Whether to validate the response against the expected schema.

    > [dYdX API docs](https://docs.dydx.xyz/indexer-client/http#get-funding-payments)
    """
    page = 1
    while True:
      response = await self.get_funding_payments(
        address=address,
        subaccount_number=subaccount_number,
        ticker=ticker,
        after_or_at=after_or_at,
        limit=limit,
        page=page,
        validate=validate,
      )
      funding_payments = response['fundingPayments']
      if not funding_payments:
        break
      yield funding_payments
      page += 1



from typing_extensions import AsyncIterable, Literal, NotRequired, TypedDict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
import pydantic

from typed_core.ws import Stream
from ..core import StreamsMixin, Unsubscribed

class OraclePriceMarket(TypedDict):
  oraclePrice: Decimal
  effectiveAt: datetime
  effectiveAtHeight: str
  marketId: int

class PerpetualMarket(TypedDict):
  clobPairId: str
  ticker: str
  status: Literal['ACTIVE', 'PAUSED', 'CANCEL_ONLY', 'POST_ONLY', 'INITIALIZING', 'FINAL_SETTLEMENT']
  oraclePrice: NotRequired[Decimal | None]
  priceChange24H: Decimal
  volume24H: Decimal
  trades24H: int
  nextFundingRate: Decimal
  initialMarginFraction: Decimal
  maintenanceMarginFraction: Decimal
  openInterest: Decimal
  atomicResolution: int
  quantumConversionExponent: int
  tickSize: Decimal
  stepSize: Decimal
  stepBaseQuantums: int
  subticksPerTick: int
  marketType: Literal['CROSS', 'ISOLATED']
  openInterestLowerCap: NotRequired[Decimal | None]
  openInterestUpperCap: NotRequired[Decimal | None]
  baseOpenInterest: Decimal
  defaultFundingRate1H: NotRequired[Decimal | None]

class TradingPerpetualMarket(TypedDict):
  atomicResolution: NotRequired[int|None]
  baseAsset: NotRequired[str|None]
  baseOpenInterest: NotRequired[Decimal|None]
  basePositionSize: NotRequired[Decimal|None]
  clobPairId: NotRequired[str|None]
  id: NotRequired[str|None]
  incrementalPositionSize: NotRequired[Decimal|None]
  initialMarginFraction: NotRequired[Decimal|None]
  maintenanceMarginFraction: NotRequired[Decimal|None]
  marketId: NotRequired[int|None]
  marketType: NotRequired[Literal['PERPETUAL', 'SPOT']|None]
  maxPositionSize: NotRequired[Decimal|None]
  nextFundingRate: NotRequired[Decimal|None]
  openInterest: NotRequired[Decimal|None]
  oraclePrice: NotRequired[Decimal | None]
  priceChange24H: NotRequired[Decimal|None]
  quantumConversionExponent: NotRequired[int|None]
  quoteAsset: NotRequired[str|None]
  status: NotRequired[Literal['ACTIVE', 'PAUSED', 'CANCEL_ONLY', 'POST_ONLY', 'INITIALIZING', 'FINAL_SETTLEMENT']|None]
  stepBaseQuantums: NotRequired[int|None]
  stepSize: NotRequired[Decimal|None]
  subticksPerTick: NotRequired[int|None]
  tickSize: NotRequired[Decimal|None]
  ticker: NotRequired[str|None]
  trades24H: NotRequired[int|None]
  volume24H: NotRequired[Decimal|None]

class Reply(TypedDict):
  markets: dict[str, PerpetualMarket]

class Notification(TypedDict):
  trading: NotRequired[dict[str, TradingPerpetualMarket]|None]
  oraclePrices: NotRequired[dict[str, OraclePriceMarket]|None]

reply_adapter = pydantic.TypeAdapter(Reply)
notification_adapter = pydantic.TypeAdapter(Notification)

@dataclass
class Markets(StreamsMixin):
  async def markets(
    self, *, batched: bool = True, validate: bool | None = None,
  ) -> Stream[Notification, Reply, Unsubscribed]:
    """
    Subscribe to the indexer markets feed.

    - `batched`: Reduce incoming messages by batching contents.
    - `validate`: Whether to validate reply and update payloads against the generated schemas.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/websockets#markets)
    """
    stream = await self.client.subscribe('v4_markets', {'batched': batched})

    async def parsed_stream() -> AsyncIterable[Notification]:
      async for msg in stream:
        data = msg['contents']
        msgs: list[Notification] = data if batched else [data]
        for d in msgs:
          yield notification_adapter.validate_python(d) if self.validate(validate) else d

    c = stream.reply['contents']
    reply: Reply = reply_adapter.validate_python(c) if self.validate(validate) else c
    return Stream(reply, parsed_stream(), stream.unsubscribe)




from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import TypedDict
from ..core import IndexerMixin, response_parser

class PnlTick(TypedDict):
  blockHeight: str
  blockTime: datetime
  createdAt: datetime
  equity: Decimal
  totalPnl: Decimal
  netTransfer: Decimal

parse_response = response_parser(list[PnlTick])

@dataclass
class GetHistoricalPnl(IndexerMixin):
  async def get_historical_pnl(
    self,
    *,
    address: str,
    subaccount_number: int,
    limit: int | None = None,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    created_on_or_after_height: int | None = None,
    created_on_or_after: datetime | None = None,
    page: int | None = None,
    validate: bool | None = None
  ) -> list[PnlTick]:
    """
    Get historical pnl

    - `address`: Wallet address that owns the account.
    - `subaccount_number`: Subaccount number.
    - `limit`: Maximum number of results to return.
    - `created_before_or_at_height`: Restrict results to entries created at or before a specific block height.
    - `created_before_or_at`: Restrict results to entries created at or before a specific timestamp.
    - `created_on_or_after_height`: Restrict results to entries created on or after a specific block height.
    - `created_on_or_after`: Restrict results to entries created on or after a specific timestamp.
    - `page`: Page number for paginated results.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-historical-pnl)
    """
    params: dict[str, object] = {
      'address': address,
      'subaccountNumber': subaccount_number,
    }
    if limit is not None:
      params['limit'] = limit
    if created_before_or_at_height is not None:
      params['createdBeforeOrAtHeight'] = created_before_or_at_height
    if created_before_or_at is not None:
      params['createdBeforeOrAt'] = ts.dump(created_before_or_at)
    if created_on_or_after_height is not None:
      params['createdOnOrAfterHeight'] = created_on_or_after_height
    if created_on_or_after is not None:
      params['createdOnOrAfter'] = ts.dump(created_on_or_after)
    if page is not None:
      params['page'] = page
    r = await self.request('GET', '/v4/historical-pnl', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass

from .api.list_positions import ListPositions, PerpetualPosition


@dataclass
class GetOpenPosition(ListPositions):
  async def get_open_position(
    self,
    market: str,
    *,
    address: str,
    subaccount: int,
    validate: bool | None = None,
  ) -> PerpetualPosition | None:
    """Retrieves the open perpetual position for a specific subaccount."""
    response = await self.list_positions(
      address,
      subaccount=subaccount,
      status='OPEN',
      validate=validate,
    )
    positions = response['positions']
    for position in positions:
      if position['market'] == market:
        return position
    return None



from typing_extensions import Literal
from dataclasses import dataclass

from .api.candles import Candles as CandlesAPI

Resolution = Literal['1MIN', '5MINS', '15MINS', '30MINS', '1HOUR', '4HOURS', '1DAY']


@dataclass
class Candles(CandlesAPI):
  async def candles(
    self,
    market: str,
    *,
    resolution: Resolution,
    validate: bool | None = None,
    batched: bool = True,
  ):
    """Subscribe to candle updates using market and resolution."""
    return await self.raw_candles(
      id=f'{market}/{resolution}',
      batched=batched,
      validate=validate,
    )



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing_extensions import Literal, TypedDict
from ..core import IndexerMixin, response_parser

class Trade(TypedDict):
  id: str
  side: Literal['BUY', 'SELL']
  size: Decimal
  price: Decimal
  type: Literal['LIMIT', 'LIQUIDATED', 'DELEVERAGED']
  createdAt: datetime
  createdAtHeight: str

class TradesResponse(TypedDict):
  trades: list[Trade]

parse_response = response_parser(TradesResponse)

@dataclass
class GetTrades(IndexerMixin):
  async def get_trades(
    self,
    market: str,
    *,
    starting_before_or_at_height: int | None = None,
    limit: int | None = None,
    validate: bool | None = None
  ) -> TradesResponse:
    """
    Retrieve trades for a perpetual market.

    - `market`: Perpetual market ticker.
    - `starting_before_or_at_height`: Latest block height to include.
    - `limit`: Maximum number of trades to return.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-trades)
    """
    params: dict[str, object] = {}
    if starting_before_or_at_height is not None:
      params['startingBeforeOrAtHeight'] = starting_before_or_at_height
    if limit is not None:
      params['limit'] = limit
    r = await self.request('GET', f'/v4/trades/perpetualMarket/{market}', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass

from .core import INDEXER_WS_URL
from .api.block_height import BlockHeight
from .candles import Candles
from .api.markets import Markets
from .api.orders import Orders
from .parent_subaccounts import ParentSubaccounts
from .subaccounts import Subaccounts
from .api.trades import Trades

@dataclass
class IndexerStreams(
  BlockHeight,
  Candles,
  Markets,
  Orders,
  ParentSubaccounts,
  Subaccounts,
  Trades,
):
  ...



from dataclasses import dataclass

from .core import INDEXER_HTTP_URL
from .api.get_asset_positions import GetAssetPositions
from .get_candles_paged import GetCandlesPaged
from .api.get_compliance_screen import GetComplianceScreen
from .get_fills_paged import GetFillsPaged
from .get_funding_payments_paged import GetFundingPaymentsPaged
from .api.get_funding_payments_for_parent_subaccount import GetFundingPaymentsForParentSubaccount
from .api.get_height import GetHeight
from .get_historical_funding_paged import GetHistoricalFundingPaged
from .api.get_historical_pnl import GetHistoricalPnl
from .get_market import GetMarket
from .api.get_megavault_historical_pnl import GetMegavaultHistoricalPnl
from .api.get_megavault_positions import GetMegavaultPositions
from .api.get_order import GetOrder
from .api.get_order_book import GetOrderBook
from .api.get_parent_asset_positions import GetParentAssetPositions
from .api.get_parent_fills import GetParentFills
from .api.get_parent_historical_pnl import GetParentHistoricalPnl
from .api.get_parent_subaccount import GetParentSubaccount
from .api.get_parent_subaccount_number_fills import GetParentSubaccountNumberFills
from .api.get_parent_subaccount_number_orders import GetParentSubaccountNumberOrders
from .api.get_parent_subaccount_number_transfers import GetParentSubaccountNumberTransfers
from .api.get_parent_transfers import GetParentTransfers
from .api.get_rewards import GetRewards
from .api.get_rewards_aggregated import GetRewardsAggregated
from .api.get_screen import GetScreen
from .api.get_sparklines import GetSparklines
from .api.get_subaccount import GetSubaccount
from .api.get_subaccounts import GetSubaccounts
from .api.get_time import GetTime
from .api.get_trades import GetTrades
from .get_transfers_paged import GetTransfersPaged
from .api.get_transfers_between import GetTransfersBetween
from .api.get_vaults_historical_pnl import GetVaultsHistoricalPnl
from .api.list_orders import ListOrders
from .api.list_parent_orders import ListParentOrders
from .api.list_parent_positions import ListParentPositions
from .get_open_position import GetOpenPosition

@dataclass
class IndexerData(
  GetAssetPositions,
  GetCandlesPaged,
  GetComplianceScreen,
  GetFillsPaged,
  GetFundingPaymentsPaged,
  GetFundingPaymentsForParentSubaccount,
  GetHeight,
  GetHistoricalFundingPaged,
  GetHistoricalPnl,
  GetMarket,
  GetMegavaultHistoricalPnl,
  GetMegavaultPositions,
  GetOrder,
  GetOrderBook,
  GetParentAssetPositions,
  GetParentFills,
  GetParentHistoricalPnl,
  GetParentSubaccount,
  GetParentSubaccountNumberFills,
  GetParentSubaccountNumberOrders,
  GetParentSubaccountNumberTransfers,
  GetParentTransfers,
  GetRewards,
  GetRewardsAggregated,
  GetScreen,
  GetSparklines,
  GetSubaccount,
  GetSubaccounts,
  GetTime,
  GetTrades,
  GetTransfersPaged,
  GetTransfersBetween,
  GetVaultsHistoricalPnl,
  ListOrders,
  ListParentOrders,
  ListParentPositions,
  GetOpenPosition,
):
  ...




from typing_extensions import Literal, TypedDict, NotRequired, Any, cast
from dataclasses import dataclass
from decimal import Decimal

from v4_proto.dydxprotocol.clob.tx_pb2 import MsgPlaceOrder
from v4_proto.dydxprotocol.clob.order_pb2 import Order as OrderProto
from v4_proto.cosmos.tx.v1beta1.service_pb2 import BroadcastTxResponse, BroadcastMode
from dydx_v4_client import OrderFlags
from dydx_v4_client.node.market import Market
from dydx_v4_client.indexer.rest.constants import OrderType
from dydx_v4_client.node.builder import TxOptions

from dydx.core import ApiError, SHORT_BLOCK_WINDOW, STATEFUL_ORDER_TIME_WINDOW
from dydx.indexer.types import PerpetualMarket
from dydx.node.core import PrivateNodeMixin

Side = Literal['BUY', 'SELL']
TimeInForce = Literal['GOOD_TIL_TIME', 'IMMEDIATE_OR_CANCEL', 'POST_ONLY', 'FILL_OR_KILL']
ConditionType = Literal['TAKE_PROFIT', 'STOP_LOSS']
Flags = Literal['LONG_TERM', 'SHORT_TERM', 'CONDITIONAL']

class Order(TypedDict):
  side: Side
  size: Decimal
  price: Decimal
  flags: Flags
  time_in_force: NotRequired[TimeInForce]
  client_metadata: NotRequired[int]
  condition_type: NotRequired[ConditionType]
  conditional_order_trigger_subticks: NotRequired[int]
  client_id: NotRequired[int]
  reduce_only: NotRequired[bool]

def rand_id() -> int:
  import random
  return random.randint(0, 100000000)

def parse_flags(flags: Flags) -> OrderFlags:
  match flags:
    case 'LONG_TERM':
      return OrderFlags.LONG_TERM
    case 'SHORT_TERM':
      return OrderFlags.SHORT_TERM
    case 'CONDITIONAL':
      return OrderFlags.CONDITIONAL

def parse_type(type: Literal['MARKET', 'LIMIT']) -> Any:
  match type:
    case 'MARKET':
      return OrderType.MARKET
    case 'LIMIT':
      return OrderType.LIMIT

def parse_side(side: Side) -> OrderProto.Side:
  match side:
    case 'BUY':
      return OrderProto.Side.SIDE_BUY
    case 'SELL':
      return OrderProto.Side.SIDE_SELL

def parse_tif(tif: TimeInForce | None) -> OrderProto.TimeInForce:
  match tif:
    case 'IMMEDIATE_OR_CANCEL':
      return OrderProto.TimeInForce.TIME_IN_FORCE_IOC
    case 'POST_ONLY':
      return OrderProto.TimeInForce.TIME_IN_FORCE_POST_ONLY
    case 'FILL_OR_KILL':
      return OrderProto.TimeInForce.TIME_IN_FORCE_FILL_OR_KILL
    case _:
      return OrderProto.TimeInForce.TIME_IN_FORCE_UNSPECIFIED

class OrderResponse(TypedDict):
  tx: BroadcastTxResponse
  order: OrderProto

@dataclass
class PlaceOrder(PrivateNodeMixin):

  def build_order(
    self, *, market: PerpetualMarket, order: Order,
    good_til_block: int | None = None, good_til_block_time: int | None = None,
    subaccount: int = 0,
  ):
    mkt = Market(market) # type: ignore
    client_id = order.get('client_id') or rand_id()
    order_id = mkt.order_id(
      address=self.address, subaccount_number=subaccount,
      client_id=client_id, order_flags=parse_flags(order['flags']),
    )
    return OrderProto(
      order_id=order_id,
      side=parse_side(order['side']),
      quantums=mkt.calculate_quantums(float(order['size'])),
      subticks=mkt.calculate_subticks(float(order['price'])),
      good_til_block=good_til_block,
      good_til_block_time=good_til_block_time,
      time_in_force=parse_tif(order.get('time_in_force')),
      reduce_only=order.get('reduce_only') or False,
      client_metadata=order.get('client_metadata'),
      condition_type=order.get('condition_type'),
      conditional_order_trigger_subticks=order.get('conditional_order_trigger_subticks'),
    )

  async def build_order_now(
    self, market: PerpetualMarket, order: Order, *, subaccount: int = 0,
    gtb_delta: int | None = None, gtbt_delta: int | None = None
  ):
    if (gtb := order.get('good_til_block')) is None and order['flags'] == 'SHORT_TERM':
      latest_block = await self.node_client.latest_block()
      if gtb_delta is None:
        gtb_delta = SHORT_BLOCK_WINDOW
      gtb = latest_block.block.header.height + gtb_delta

    if (gtbt := order.get('good_til_block_time')) is None and order['flags'] == 'LONG_TERM':
      latest_block = await self.node_client.latest_block()
      if gtbt_delta is None:
        gtbt_delta = STATEFUL_ORDER_TIME_WINDOW
      gtbt = latest_block.block.header.time.seconds + gtbt_delta
      gtb = None

    return self.build_order(market=market, order=order, good_til_block=gtb, good_til_block_time=gtbt, subaccount=subaccount)

  async def place_order(
    self, market: PerpetualMarket, order: Order, *, subaccount: int = 0,
    mode: BroadcastMode = BroadcastMode.BROADCAST_MODE_SYNC, tx_options: TxOptions | None = None,
    gtb_delta: int | None = None, gtbt_delta: int | None = None,
    good_til_block: int | None = None, good_til_block_time: int | None = None,
  ) -> OrderResponse:
    """Place an order.
    
    - `market`: market to place the order on
    - `order`: order to place
    - `subaccount`: subaccount to place the order on
    - `gtb_delta`: sets good-til-block for short-term orders to `current_block() + gtb_delta`
    - `gtbt_delta`: sets good-til-block-time (in seconds) for long-term-orders to  `current_block().time.seconds + gtbt_delta`

    > [dYdX API docs](https://docs.dydx.xyz/node-client/private#place-order)
    """
    if good_til_block is not None or good_til_block_time is not None:
      order_proto = self.build_order(market=market, order=order, good_til_block=good_til_block, good_til_block_time=good_til_block_time, subaccount=subaccount)
    else:
      order_proto = await self.build_order_now(market=market, order=order, subaccount=subaccount, gtb_delta=gtb_delta, gtbt_delta=gtbt_delta)
    tx: BroadcastTxResponse = await self.node_client.broadcast_message(
      self.wallet, MsgPlaceOrder(order=order_proto), # type: ignore
      mode=mode, tx_options=tx_options
    )
    if tx.tx_response.code != 0:
      raise ApiError(tx.tx_response.code, tx.tx_response, tx.tx_response.raw_log)
    return {
      'tx': tx,
      'order': order_proto,
    }


from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class Account(TypedDict):
  address: str
  subaccountNumber: NotRequired[int|None]

class Transfer(TypedDict):
  id: str
  sender: Account
  recipient: Account
  size: Decimal
  createdAt: datetime
  createdAtHeight: str
  symbol: str
  type: Literal['TRANSFER_IN', 'TRANSFER_OUT', 'DEPOSIT', 'WITHDRAWAL']
  transactionHash: str

class TransfersResponse(TypedDict):
  transfers: list[Transfer]

parse_response = response_parser(TransfersResponse)

@dataclass
class GetTransfersBetween(IndexerMixin):
  async def get_transfers_between(
    self,
    *,
    source_address: str,
    source_subaccount_number: str,
    recipient_address: str,
    recipient_subaccount_number: str,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    validate: bool | None = None
  ) -> TransfersResponse:
    """
    Get transfers between subaccounts

    - `source_address`: Sender wallet address.
    - `source_subaccount_number`: Sender subaccount number.
    - `recipient_address`: Recipient wallet address.
    - `recipient_subaccount_number`: Recipient subaccount number.
    - `created_before_or_at_height`: Restrict results to entries created at or before a specific block height.
    - `created_before_or_at`: Restrict results to entries created at or before a specific timestamp.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-transfers-between)
    """
    params: dict[str, object] = {
      'sourceAddress': source_address,
      'sourceSubaccountNumber': source_subaccount_number,
      'recipientAddress': recipient_address,
      'recipientSubaccountNumber': recipient_subaccount_number,
    }
    if created_before_or_at_height is not None:
      params['createdBeforeOrAtHeight'] = created_before_or_at_height
    if created_before_or_at is not None:
      params['createdBeforeOrAt'] = ts.dump(created_before_or_at)
    r = await self.request('GET', '/v4/transfers/between', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass
from decimal import Decimal

from grpc._channel import _InactiveRpcError

from dydx.core import ApiError
from dydx.node.core import PublicNodeMixin

@dataclass
class FeeTier:
  maker: Decimal
  """Maker fee (relative to base fill size, e.g. 0.01 = 1%)"""
  taker: Decimal
  """Taker fee (relative to base fill size, e.g. 0.01 = 1%)"""

@dataclass
class GetUserFeeTier(PublicNodeMixin):
  async def get_user_fee_tier(self, address: str) -> FeeTier:
    """
    Retrieves the perpetual fee tier associated with a specific wallet address, providing information on the user's current fee structure.

    - `address`: The wallet address that owns the account.

    > [dYdX API docs](https://docs.dydx.xyz/node-client/public#get-fee-tiers)
    """
    try:
      r = await self.node_client.get_user_fee_tier(address)
      return FeeTier(
        maker=Decimal(r.tier.maker_fee_ppm) / Decimal('1e6'),
        taker=Decimal(r.tier.taker_fee_ppm) / Decimal('1e6'),
      )
    except _InactiveRpcError as e:
      raise ApiError(e._state.code, e._state.details)


from typing_extensions import Literal

PositionSide = Literal['LONG', 'SHORT']
OrderSide = Literal['BUY', 'SELL']

OrderStatus = Literal[
  'OPEN',
  'FILLED',
  'CANCELED',
  'BEST_EFFORT_CANCELED',
  'UNTRIGGERED',
  'BEST_EFFORT_OPENED',
  'PENDING',
]

OrderType = Literal[
  'LIMIT',
  'MARKET',
  'STOP_LIMIT',
  'STOP_MARKET',
  'TRAILING_STOP',
  'TAKE_PROFIT',
  'TAKE_PROFIT_MARKET',
  'HARD_TRADE',
  'FAILED_HARD_TRADE',
  'TRANSFER_PLACEHOLDER',
]
"""
`LIMIT`: Executes at a specified price or better.
`MARKET`: Executes immediately at the best available price.
`STOP_LIMIT`: Becomes a limit order once a stop price is reached.
`STOP_MARKET`: Becomes a market order once a stop price is reached.
`TRAILING_STOP`: A dynamic stop order that adjusts based on market movement.
`TAKE_PROFIT`: A limit order to secure profits once a target price is reached.
`TAKE_PROFIT_MARKET`: A market order to secure profits once a target price is reached.
`HARD_TRADE`: A special internal trade type, typically used in matching engines.
`FAILED_HARD_TRADE`: Represents a failed execution of a HardTrade.
`TRANSFER_PLACEHOLDER`: A placeholder type used internally for transfers.
"""

TimeInForce = Literal['GTT', 'IOC', 'FOK']
Liquidity = Literal['MAKER', 'TAKER']
FillType = Literal['LIMIT', 'LIQUIDATED', 'LIQUIDATION', 'DELEVERAGED', 'OFFSETTING']
"""
`LIMIT`: Result of a regular limit order.
`LIQUIDATED`: Result of liquidating another trader's position.
`LIQUIDATION`: Result of one's own position being liquidated.
`DELEVERAGED`: Caused by automatic deleveraging in risk management.
`OFFSETTING`: Result of offsetting positions, often for risk reduction.
"""

TransferType = Literal['TRANSFER_IN', 'TRANSFER_OUT', 'DEPOSIT', 'WITHDRAWAL']


from .enums import (
  PositionSide, OrderSide, OrderStatus, OrderType, TimeInForce,
  Liquidity, FillType, TransferType,
)
from .order import OrderState
from .positions import PerpetualPosition, PerpetualPositionStatus, AssetPosition
from .misc import Account, Subaccount
from .market import PerpetualMarket

__all__ = [
  'PositionSide', 'OrderSide', 'OrderStatus', 'OrderType', 'TimeInForce',
  'Liquidity', 'FillType', 'TransferType',
  'OrderState',
  'PerpetualPosition', 'PerpetualPositionStatus', 'AssetPosition',
  'Account', 'Subaccount', 'PerpetualMarket',
]


from typing_extensions import AsyncIterable, Literal, NotRequired, TypedDict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
import pydantic

from typed_core.ws import Stream
from ..core import StreamsMixin, Unsubscribed

class Candle(TypedDict):
  startedAt: datetime
  ticker: str
  resolution: Literal['1MIN', '5MINS', '15MINS', '30MINS', '1HOUR', '4HOURS', '1DAY']
  low: Decimal
  high: Decimal
  open: Decimal
  close: Decimal
  baseTokenVolume: Decimal
  usdVolume: Decimal
  trades: int
  startingOpenInterest: Decimal
  orderbookMidPriceOpen: NotRequired[Decimal | None]
  orderbookMidPriceClose: NotRequired[Decimal | None]

class Notification(TypedDict):
  startedAt: datetime
  ticker: str
  resolution: Literal['1MIN', '5MINS', '15MINS', '30MINS', '1HOUR', '4HOURS', '1DAY']
  low: Decimal
  high: Decimal
  open: Decimal
  close: Decimal
  baseTokenVolume: Decimal
  usdVolume: Decimal
  trades: int
  startingOpenInterest: Decimal
  orderbookMidPriceOpen: NotRequired[Decimal | None]
  orderbookMidPriceClose: NotRequired[Decimal | None]

class Reply(TypedDict):
  candles: list[Candle]

reply_adapter = pydantic.TypeAdapter(Reply)
notification_adapter = pydantic.TypeAdapter(Notification)

@dataclass
class Candles(StreamsMixin):
  async def raw_candles(
    self, *, id: str, batched: bool = True, validate: bool | None = None,
  ) -> Stream[Notification, Reply, Unsubscribed]:
    """
    Subscribe to the indexer candles feed for a market and resolution.

    - `id`: Market and candle resolution formatted as {market}/{resolution}.
    - `batched`: Reduce incoming messages by batching contents.
    - `validate`: Whether to validate reply and update payloads against the generated schemas.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/websockets#candles)
    """
    stream = await self.client.subscribe(f'v4_candles:{id}', {'batched': batched})

    async def parsed_stream() -> AsyncIterable[Notification]:
      async for msg in stream:
        data = msg['contents']
        msgs: list[Notification] = data if batched else [data]
        for d in msgs:
          yield notification_adapter.validate_python(d) if self.validate(validate) else d

    c = stream.reply['contents']
    reply: Reply = reply_adapter.validate_python(c) if self.validate(validate) else c
    return Stream(reply, parsed_stream(), stream.unsubscribe)




from typing_extensions import TypeVar, Mapping
from datetime import datetime, timezone
from decimal import Decimal, ROUND_HALF_DOWN, ROUND_FLOOR

D = TypeVar('D', bound=Mapping)

def filter_kwargs(Params: type[D], params: D | dict) -> D:
  return { k: params[k] for k in getattr(Params, '__annotations__', {}) if k in params } # type: ignore

class timestamp:
  @staticmethod
  def parse(time: str) -> datetime:
    return datetime.fromisoformat(time).astimezone()

  @staticmethod
  def dump(dt: datetime) -> str:
    return dt.astimezone(timezone.utc).isoformat()
  
  @staticmethod
  def now() -> str:
    return timestamp.dump(datetime.now())

def round2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).quantize(Decimal('1.'), rounding=ROUND_HALF_DOWN) * tick_size
  return r.normalize()

def trunc2tick(x: Decimal, tick_size: Decimal) -> Decimal:
  r = (x / tick_size).to_integral_value(rounding=ROUND_FLOOR) * tick_size
  return r.normalize()

def path_join(base: str, *parts: str):
  return '/'.join([base.rstrip('/')] + [part.lstrip('/') for part in parts])

def getenv(var: str) -> str:
  import os
  try:
    return os.environ[var]
  except KeyError:
    raise ValueError(f'Environment variable {var} not found')



from dataclasses import dataclass

from .api.parent_subaccounts import ParentSubaccounts as ParentSubaccountsAPI

@dataclass
class ParentSubaccounts(ParentSubaccountsAPI):
  async def parent_subaccounts(
    self, address: str, *,
    subaccount: int,
    validate: bool | None = None,
    batched: bool = True,
  ):
    """Subscribe to parent subaccount updates using address and subaccount number."""
    return await self.raw_parent_subaccounts(
      id=f'{address}/{subaccount}',
      batched=batched,
      validate=validate,
    )



from typing_extensions import AsyncIterable, Literal, NotRequired, TypedDict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
import pydantic

from typed_core.ws import Stream
from ..core import StreamsMixin, Unsubscribed

class Account(TypedDict):
  address: str
  subaccountNumber: NotRequired[int|None]

class AssetPosition(TypedDict):
  size: Decimal
  symbol: str
  side: Literal['LONG', 'SHORT']
  assetId: str
  subaccountNumber: int

class AssetPositionSubaccountMessage(TypedDict):
  address: str
  subaccountNumber: int
  positionId: str
  assetId: str
  symbol: str
  side: Literal['LONG', 'SHORT']
  size: Decimal

class FillSubaccountMessage(TypedDict):
  id: str
  subaccountId: str
  side: Literal['BUY', 'SELL']
  liquidity: Literal['MAKER', 'TAKER']
  type: Literal['LIMIT', 'LIQUIDATED', 'LIQUIDATION', 'DELEVERAGED', 'OFFSETTING']
  clobPairId: str
  size: Decimal
  price: Decimal
  quoteAmount: Decimal
  eventId: str
  transactionHash: str
  createdAt: datetime
  createdAtHeight: NotRequired[str | None]
  ticker: str
  orderId: NotRequired[str | None]
  clientMetadata: NotRequired[str | None]

class Order(TypedDict):
  id: str
  subaccountId: str
  clientId: str
  clobPairId: str
  side: Literal['BUY', 'SELL']
  size: Decimal
  totalFilled: Decimal
  price: Decimal
  type: Literal['LIMIT', 'MARKET', 'STOP_LIMIT', 'STOP_MARKET', 'TRAILING_STOP', 'TAKE_PROFIT', 'TAKE_PROFIT_MARKET', 'HARD_TRADE', 'FAILED_HARD_TRADE', 'TRANSFER_PLACEHOLDER']
  status: Literal['OPEN', 'FILLED', 'CANCELED', 'BEST_EFFORT_CANCELED', 'UNTRIGGERED', 'BEST_EFFORT_OPENED', 'PENDING']
  timeInForce: Literal['GTT', 'IOC', 'FOK']
  reduceOnly: NotRequired[bool | None]
  orderFlags: str
  goodTilBlock: NotRequired[str | None]
  goodTilBlockTime: NotRequired[datetime | None]
  createdAtHeight: NotRequired[str | None]
  clientMetadata: NotRequired[str | None]
  triggerPrice: NotRequired[Decimal | None]
  postOnly: NotRequired[bool | None]
  ticker: str
  updatedAt: NotRequired[datetime | None]
  updatedAtHeight: NotRequired[str | None]
  subaccountNumber: int
  orderRouterAddress: NotRequired[str|None]
  builderFee: NotRequired[Decimal | None]
  feePpm: NotRequired[Decimal | None]

class OrderSubaccountMessage(TypedDict):
  id: str
  subaccountId: str
  clientId: str
  clobPairId: str
  side: Literal['BUY', 'SELL']
  size: Decimal
  ticker: NotRequired[str | None]
  price: NotRequired[Decimal | None]
  type: Literal['LIMIT', 'MARKET', 'STOP_LIMIT', 'STOP_MARKET', 'TRAILING_STOP', 'TAKE_PROFIT', 'TAKE_PROFIT_MARKET', 'HARD_TRADE', 'FAILED_HARD_TRADE', 'TRANSFER_PLACEHOLDER']
  timeInForce: Literal['GTT', 'IOC', 'FOK']
  postOnly: NotRequired[bool | None]
  reduceOnly: NotRequired[bool | None]
  status: Literal['OPEN', 'FILLED', 'CANCELED', 'BEST_EFFORT_CANCELED', 'UNTRIGGERED', 'BEST_EFFORT_OPENED', 'PENDING']
  orderFlags: NotRequired[str | None]
  totalFilled: NotRequired[Decimal | None]
  totalOptimisticFilled: NotRequired[Decimal | None]
  goodTilBlock: NotRequired[str | None]
  goodTilBlockTime: NotRequired[datetime | None]
  triggerPrice: NotRequired[Decimal | None]
  updatedAt: NotRequired[datetime | None]
  updatedAtHeight: NotRequired[str | None]
  removalReason: NotRequired[str | None]
  createdAtHeight: NotRequired[str | None]
  clientMetadata: NotRequired[str | None]
  subaccountNumber: NotRequired[int | None]
  orderRouterAddress: NotRequired[str | None]

class PerpetualPosition(TypedDict):
  market: str
  status: Literal['OPEN', 'CLOSED', 'LIQUIDATED']
  side: Literal['LONG', 'SHORT']
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal | None]
  realizedPnl: NotRequired[Decimal | None]
  unrealizedPnl: NotRequired[Decimal | None]
  createdAt: datetime
  createdAtHeight: str
  closedAt: NotRequired[datetime | None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

class PerpetualPositionSubaccountMessage(TypedDict):
  market: str
  status: Literal['OPEN', 'CLOSED', 'LIQUIDATED']
  side: Literal['LONG', 'SHORT']
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal | None]
  realizedPnl: NotRequired[Decimal | None]
  unrealizedPnl: NotRequired[Decimal | None]
  createdAt: NotRequired[datetime | None]
  createdAtHeight: NotRequired[str | None]
  closedAt: NotRequired[datetime | None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

class TradingRewardSubaccountMessage(TypedDict):
  tradingReward: Decimal
  createdAt: NotRequired[datetime | None]
  createdAtHeight: NotRequired[str | None]

AssetPositionsMap = dict[str, AssetPosition]

PerpetualPositionsMap = dict[str, PerpetualPosition]

class TransferSubaccountMessage(TypedDict):
  sender: Account
  recipient: Account
  symbol: str
  size: Decimal
  type: Literal['TRANSFER_IN', 'TRANSFER_OUT', 'DEPOSIT', 'WITHDRAWAL']
  transactionHash: str
  createdAt: NotRequired[datetime | None]
  createdAtHeight: NotRequired[str | None]

class Subaccount(TypedDict):
  address: str
  subaccountNumber: int
  equity: Decimal
  freeCollateral: Decimal
  openPerpetualPositions: PerpetualPositionsMap
  assetPositions: AssetPositionsMap
  marginEnabled: bool
  updatedAtHeight: str
  latestProcessedBlockHeight: str

class Notification(TypedDict):
  perpetualPositions: NotRequired[list[PerpetualPositionSubaccountMessage]|None]
  assetPositions: NotRequired[list[AssetPositionSubaccountMessage]|None]
  orders: NotRequired[list[OrderSubaccountMessage]|None]
  fills: NotRequired[list[FillSubaccountMessage]|None]
  transfers: NotRequired[list[TransferSubaccountMessage] | TransferSubaccountMessage|None]
  tradingReward: NotRequired[TradingRewardSubaccountMessage|None]
  tradingRewards: NotRequired[list[TradingRewardSubaccountMessage]|None]
  blockHeight: NotRequired[str | None]

class Reply(TypedDict):
  subaccount: Subaccount
  orders: list[Order]
  blockHeight: str

reply_adapter = pydantic.TypeAdapter(Reply)
notification_adapter = pydantic.TypeAdapter(Notification)

@dataclass
class Subaccounts(StreamsMixin):
  async def raw_subaccounts(
    self, *, id: str, batched: bool = True, validate: bool | None = None,
  ) -> Stream[Notification, Reply, Unsubscribed]:
    """
    Subscribe to the indexer subaccounts feed for an address and subaccount number.

    - `id`: Subaccount id formatted as {address}/{subaccount-number}.
    - `batched`: Reduce incoming messages by batching contents.
    - `validate`: Whether to validate reply and update payloads against the generated schemas.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/websockets#subaccounts)
    """
    stream = await self.client.subscribe(f'v4_subaccounts:{id}', {'batched': batched})

    async def parsed_stream() -> AsyncIterable[Notification]:
      async for msg in stream:
        data = msg['contents']
        msgs: list[Notification] = data if batched else [data]
        for d in msgs:
          yield notification_adapter.validate_python(d) if self.validate(validate) else d

    c = stream.reply['contents']
    reply: Reply = reply_adapter.validate_python(c) if self.validate(validate) else c
    return Stream(reply, parsed_stream(), stream.unsubscribe)




from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class Fill(TypedDict):
  id: str
  side: Literal['BUY', 'SELL']
  liquidity: Literal['MAKER', 'TAKER']
  type: Literal['LIMIT', 'LIQUIDATED', 'LIQUIDATION', 'DELEVERAGED', 'OFFSETTING']
  market: str
  marketType: Literal['PERPETUAL', 'SPOT']
  price: Decimal
  size: Decimal
  fee: Decimal
  affiliateRevShare: Decimal
  createdAt: datetime
  createdAtHeight: str
  orderId: NotRequired[str | None]
  clientMetadata: NotRequired[str | None]
  subaccountNumber: int
  builderFee: NotRequired[Decimal | None]
  builderAddress: NotRequired[str | None]
  positionSizeBefore: NotRequired[Decimal | None]
  entryPriceBefore: NotRequired[Decimal | None]
  positionSideBefore: NotRequired[Literal['LONG', 'SHORT'] | None]

class FillsResponse(TypedDict):
  fills: list[Fill]

parse_response = response_parser(FillsResponse)

@dataclass
class GetParentSubaccountNumberFills(IndexerMixin):
  async def get_parent_subaccount_number_fills(
    self,
    *,
    address: str,
    parent_subaccount_number: int,
    ticker: str | None = None,
    ticker_type: Literal['PERPETUAL', 'SPOT'] | None = None,
    limit: int | None = None,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    page: int | None = None,
    validate: bool | None = None
  ) -> FillsResponse:
    """
    Get parent subaccount number fills

    - `address`: Wallet address that owns the account.
    - `parent_subaccount_number`: Parent subaccount number.
    - `ticker`: Ticker filter.
    - `ticker_type`: Ticker type filter.
    - `limit`: Maximum number of results to return.
    - `created_before_or_at_height`: Restrict results to entries created at or before a specific block height.
    - `created_before_or_at`: Restrict results to entries created at or before a specific timestamp.
    - `page`: Page number for paginated results.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/accounts#get-parent-subaccount-number-fills)
    """
    params: dict[str, object] = {
      'address': address,
      'parentSubaccountNumber': parent_subaccount_number,
    }
    if ticker is not None:
      params['ticker'] = ticker
    if ticker_type is not None:
      params['tickerType'] = ticker_type
    if limit is not None:
      params['limit'] = limit
    if created_before_or_at_height is not None:
      params['createdBeforeOrAtHeight'] = created_before_or_at_height
    if created_before_or_at is not None:
      params['createdBeforeOrAt'] = ts.dump(created_before_or_at)
    if page is not None:
      params['page'] = page
    r = await self.request('GET', '/v4/fills/parentSubaccountNumber', params=params)
    return parse_response(r, validate=self.validate(validate))




SHORT_BLOCK_WINDOW = 20
"""Max. number of blocks for a short-term order to live. See [source](https://docs.dydx.xyz/concepts/trading/orders#comparison)"""

STATEFUL_ORDER_TIME_WINDOW = 95*3600*24
"""Max. time window for a stateful order to live. See [source](https://docs.dydx.xyz/concepts/trading/orders#comparison)"""


from typing_extensions import NotRequired, Literal
from decimal import Decimal

from dydx.core import TypedDict
from .enums import PositionSide

PerpetualPositionStatus = Literal['OPEN', 'CLOSED', 'LIQUIDATED']
"""Perpetual Position Status

> [dYdX API docs](https://docs.dydx.xyz/types/perpetual_position_status)
"""

class PerpetualPosition(TypedDict):
  """Perpetual Position Response Object

  > [dYdX API docs](https://docs.dydx.xyz/types/perpetual_position_response_object)
  """
  market: str
  status: PerpetualPositionStatus
  side: PositionSide
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal|None]
  realizedPnl: NotRequired[Decimal|None]
  unrealizedPnl: NotRequired[Decimal|None]
  createdAt: str
  createdAtHeight: str
  closedAt: NotRequired[str|None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

class AssetPosition(TypedDict):
  """Asset Position Response Object

  > [dYdX API docs](https://docs.dydx.xyz/types/asset_position_response_object)
  """
  size: str
  symbol: str
  side: PositionSide
  assetId: str
  subaccountNumber: int


from dataclasses import dataclass
from typing_extensions import TypedDict
from ..core import IndexerMixin, response_parser

class ComplianceResponse(TypedDict):
  restricted: bool
  reason: str

parse_response = response_parser(ComplianceResponse)

@dataclass
class GetScreen(IndexerMixin):
  async def get_screen(
    self,
    *,
    address: str,
    validate: bool | None = None
  ) -> ComplianceResponse:
    """
    Get screen

    - `address`: Wallet address to screen.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/utility#get-screen)
    """
    params: dict[str, object] = {
      'address': address,
    }
    r = await self.request('GET', '/v4/screen', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import TypedDict
from ..core import IndexerMixin, response_parser

class Funding(TypedDict):
  ticker: str
  rate: Decimal
  price: Decimal
  effectiveAt: datetime
  effectiveAtHeight: str

class HistoricalFundingResponse(TypedDict):
  historicalFunding: list[Funding]

parse_response = response_parser(HistoricalFundingResponse)

@dataclass
class GetHistoricalFunding(IndexerMixin):
  async def get_historical_funding(
    self,
    market: str,
    *,
    effective_before_or_at: datetime | None = None,
    effective_before_or_at_height: int | None = None,
    limit: int | None = None,
    validate: bool | None = None
  ) -> HistoricalFundingResponse:
    """
    Retrieve historical funding data for a perpetual market.

    - `market`: Perpetual market ticker.
    - `effective_before_or_at`: Latest effective timestamp to include.
    - `effective_before_or_at_height`: Latest effective block height to include.
    - `limit`: Maximum number of funding entries to return.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-historical-funding)
    """
    params: dict[str, object] = {}
    if effective_before_or_at is not None:
      params['effectiveBeforeOrAt'] = ts.dump(effective_before_or_at)
    if effective_before_or_at_height is not None:
      params['effectiveBeforeOrAtHeight'] = effective_before_or_at_height
    if limit is not None:
      params['limit'] = limit
    r = await self.request('GET', f'/v4/historicalFunding/{market}', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import Any, TypeVar, Generic, Literal, TypedDict, AsyncIterable
from abc import abstractmethod
from dataclasses import dataclass, field
import asyncio

from dydx.core import UserError
from .base import RpcSocketClient

M = TypeVar('M', default=Any)
R = TypeVar('R', default=Any)
S = TypeVar('S', default=Any)
D = TypeVar('D', default=Any)

class Response(TypedDict, Generic[R]):
  kind: Literal['response']
  response: R

class Subscription(TypedDict, Generic[D]):
  kind: Literal['subscription']
  channel: str
  data: D

Message = Response[R] | Subscription[D]

@dataclass
class StreamsRPCSocketClient(RpcSocketClient[M, R], Generic[M, R, S, D]):
  lock: asyncio.Lock = field(default_factory=asyncio.Lock, init=False, repr=False)
  replies: asyncio.Queue[R] = field(default_factory=asyncio.Queue, init=False, repr=False)
  subscribers: dict[str, asyncio.Queue[D]] = field(default_factory=dict, init=False, repr=False)

  @abstractmethod
  async def req_subscription(self, channel: str, *, id: str | None = None, batched: bool | None = None) -> S:
    ...

  @abstractmethod
  async def req_unsubscription(self, channel: str):
    ...

  @abstractmethod
  def parse_msg(self, msg: str | bytes) -> Message[R, D] | None:
    ...

  def on_msg(self, msg: str | bytes):
    obj = self.parse_msg(msg)
    if obj is None:
      return
    elif obj['kind'] == 'response':
      self.replies.put_nowait(obj['response'])
    elif obj['kind'] == 'subscription':
      if (q := self.subscribers.get(obj['channel'])) is not None:
        q.put_nowait(obj['data'])

  @abstractmethod
  async def send(self, msg: M):
    ...

  async def request(self, msg: M) -> R:
    async with self.lock:
      await self.send(msg)
      return await self.replies.get()

  async def subscribe(self, channel: str, *, id: str | None = None, batched: bool | None = None) -> tuple[S, AsyncIterable[D]]:
    if channel in self.subscribers:
      raise UserError(f'Channel {channel} already subscribed')

    self.subscribers[channel] = asyncio.Queue()
    r = await self.req_subscription(channel, id=id, batched=batched)

    async def gen():
      while True:
        if (q := self.subscribers.get(channel)) is None:
          break # channel unsubscribed
        yield await self.wait_with_listener(q.get())

    return r, gen()

  async def unsubscribe(self, channel: str):
    if channel not in self.subscribers:
      raise UserError(f'Channel {channel} not subscribed')
    del self.subscribers[channel]
    await self.req_unsubscription(channel)

    


from typing_extensions import Any, TypeVar, Generic
from abc import abstractmethod
from dataclasses import dataclass, field
import asyncio

from .base import RpcSocketClient

M = TypeVar('M', default=Any)
R = TypeVar('R', default=Any)

@dataclass
class MultiplexRpcSocketClient(RpcSocketClient[M, R], Generic[M, R]):
  """Multiplexed request/response socket client. It uses IDs to identify requests and responses."""
  replies: dict[int, asyncio.Future[R]] = field(default_factory=dict)
  counter: int = 0

  @abstractmethod
  def parse_response(self, msg: str | bytes) -> tuple[int, R]:
    ...

  def on_msg(self, msg: str | bytes):
    id, result = self.parse_response(msg)
    self.replies[id].set_result(result)

  @abstractmethod
  async def send(self, id: int, msg: M):
    ...

  async def request(self, msg: M) -> R:
    id = self.counter
    self.counter += 1
    self.replies[id] = asyncio.Future()
    await self.send(id, msg)
    res = await self.wait_with_listener(self.replies[id])
    del self.replies[id]
    return res



from dataclasses import dataclass
from decimal import Decimal
from typing_extensions import TypedDict
from ..core import IndexerMixin, response_parser

class BookEntry(TypedDict):
  price: Decimal
  size: Decimal

class OrderBook(TypedDict):
  bids: list[BookEntry]
  asks: list[BookEntry]

parse_response = response_parser(OrderBook)

@dataclass
class GetOrderBook(IndexerMixin):
  async def get_order_book(
    self,
    market: str,
    validate: bool | None = None
  ) -> OrderBook:
    """
    Retrieve the orderbook for a perpetual market.

    - `market`: Perpetual market ticker.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#get-perpetual-market-orderbook)
    """
    r = await self.request('GET', f'/v4/orderbooks/perpetualMarket/{market}')
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass
from datetime import datetime
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class ComplianceV2Response(TypedDict):
  status: Literal['COMPLIANT', 'FIRST_STRIKE_CLOSE_ONLY', 'FIRST_STRIKE', 'CLOSE_ONLY', 'BLOCKED']
  reason: NotRequired[Literal['MANUAL', 'US_GEO', 'CA_GEO', 'GB_GEO', 'SANCTIONED_GEO', 'COMPLIANCE_PROVIDER'] | None]
  updatedAt: NotRequired[datetime | None]

parse_response = response_parser(ComplianceV2Response)

@dataclass
class GetComplianceScreen(IndexerMixin):
  async def get_compliance_screen(
    self,
    address: str,
    validate: bool | None = None
  ) -> ComplianceV2Response:
    """
    Get compliance screen

    - `address`: EVM or dYdX address.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/utility#get-compliance-screen)
    """
    r = await self.request('GET', f'/v4/compliance/screen/{address}')
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass

from .api.get_markets import GetMarkets, PerpetualMarket


@dataclass
class GetMarket(GetMarkets):
  async def get_market(
    self,
    market: str,
    *,
    validate: bool | None = None,
  ) -> PerpetualMarket:
    """Retrieves a single perpetual market by ticker."""
    response = await self.get_markets(market=market, limit=1, validate=validate)
    return response['markets'][market] if 'markets' in response else response[market]



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from dydx.core import timestamp as ts
from typing_extensions import Literal, NotRequired, TypedDict
from ..core import IndexerMixin, response_parser

class PerpetualPosition(TypedDict):
  market: str
  status: Literal['OPEN', 'CLOSED', 'LIQUIDATED']
  side: Literal['LONG', 'SHORT']
  size: Decimal
  maxSize: Decimal
  entryPrice: Decimal
  exitPrice: NotRequired[Decimal | None]
  realizedPnl: NotRequired[Decimal | None]
  unrealizedPnl: NotRequired[Decimal | None]
  createdAt: datetime
  createdAtHeight: str
  closedAt: NotRequired[datetime | None]
  sumOpen: Decimal
  sumClose: Decimal
  netFunding: Decimal
  subaccountNumber: int

class PositionsResponse(TypedDict):
  positions: list[PerpetualPosition]

parse_response = response_parser(PositionsResponse)

@dataclass
class ListPositions(IndexerMixin):
  async def list_positions(
    self, address: str, *,
    subaccount: int,
    status: Literal['OPEN', 'CLOSED', 'LIQUIDATED'] | None = None,
    limit: int | None = None,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    validate: bool | None = None
  ) -> PositionsResponse:
    """
    Retrieve perpetual positions for a subaccount.

    - `address`: Wallet address that owns the subaccount.
    - `subaccount`: Subaccount number.
    - `status`: Position status filter.
    - `limit`: Maximum number of positions to return.
    - `created_before_or_at_height`: Latest block height to include.
    - `created_before_or_at`: Latest timestamp to include.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http#list-positions)
    """
    params: dict[str, object] = {
      'address': address,
      'subaccountNumber': subaccount,
    }
    if status is not None:
      params['status'] = status
    if limit is not None:
      params['limit'] = limit
    if created_before_or_at_height is not None:
      params['createdBeforeOrAtHeight'] = created_before_or_at_height
    if created_before_or_at is not None:
      params['createdBeforeOrAt'] = ts.dump(created_before_or_at)
    r = await self.request('GET', '/v4/perpetualPositions', params=params)
    return parse_response(r, validate=self.validate(validate))




from dataclasses import dataclass
from decimal import Decimal
from typing_extensions import Literal
from ..core import IndexerMixin, response_parser

Sparkline = dict[str, list[Decimal]]

parse_response = response_parser(Sparkline)

@dataclass
class GetSparklines(IndexerMixin):
  async def get_sparklines(
    self,
    *,
    time_period: Literal['OneDay', 'SevenDays'],
    validate: bool | None = None
  ) -> Sparkline:
    """
    Get sparklines

    - `time_period`: Sparkline time period.
    - `validate`: Whether to validate the response against the generated schema.

    > [Official API docs](https://docs.dydx.xyz/indexer-client/http/markets#get-sparklines)
    """
    params: dict[str, object] = {
      'timePeriod': time_period,
    }
    r = await self.request('GET', '/v4/sparklines', params=params)
    return parse_response(r, validate=self.validate(validate))




from typing_extensions import AsyncIterable
from dataclasses import dataclass
from datetime import datetime

from .api.get_transfers import GetTransfers, Transfer

@dataclass
class GetTransfersPaged(GetTransfers):
  async def get_transfers_paged(
    self,
    *,
    address: str,
    subaccount_number: int,
    created_before_or_at_height: int | None = None,
    created_before_or_at: datetime | None = None,
    limit: int | None = None,
    validate: bool | None = None,
  ) -> AsyncIterable[list[Transfer]]:
    """Retrieves the transfer history for a specific subaccount.

    - `address`: The wallet address that owns the account.
    - `subaccount_number`: The identifier for the specific subaccount within the wallet address.
    - `created_before_or_at_height`: If given, fetches transfers up to and including the given block height.
    - `created_before_or_at`: If given, fetches transfers up to and including the given timestamp.
    - `limit`: The max. number of transfers to retrieve (default: 1000, max: 1000).
    - `validate`: Whether to validate the response against the expected schema.

    > [dYdX API docs](https://docs.dydx.xyz/indexer-client/http#get-transfers)
    """
    page = 1
    while True:
      response = await self.get_transfers(
        address=address,
        subaccount_number=subaccount_number,
        created_before_or_at_height=created_before_or_at_height,
        created_before_or_at=created_before_or_at,
        limit=limit,
        page=page,
        validate=validate,
      )
      transfers = response['transfers']
      if not transfers:
        break
      yield transfers
      page += 1
